#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Abstract rule class definition and rule engine implementation
"""
from abc import ABCMeta, abstractmethod
import inspect
from itertools import groupby
from logging import getLogger

from .utils import is_iterable

from .toposort import toposort

from . import debug

log = getLogger(__name__).log


class Consequence(metaclass=ABCMeta):
    """
    Definition of a consequence to apply.
    """
    @abstractmethod
    def then(self, matches, when_response, context):  # pragma: no cover
        """
        Action implementation.

        :param matches:
        :type matches: rebulk.match.Matches
        :param context:
        :type context:
        :param when_response: return object from when call.
        :type when_response: object
        :return: True if the action was runned, False if it wasn't.
        :rtype: bool
        """


class Condition(metaclass=ABCMeta):
    """
    Definition of a condition to check.
    """
    @abstractmethod
    def when(self, matches, context):  # pragma: no cover
        """
        Condition implementation.

        :param matches:
        :type matches: rebulk.match.Matches
        :param context:
        :type context:
        :return: truthy if rule should be triggered and execute then action, falsy if it should not.
        :rtype: object
        """


class CustomRule(Condition, Consequence, metaclass=ABCMeta):
    """
    Definition of a rule to apply
    """
    # pylint: disable=unused-argument, abstract-method
    priority = 0
    name = None
    dependency = None
    properties = {}

    def __init__(self, log_level=None):
        self.defined_at = debug.defined_at()
        if log_level is None and not hasattr(self, 'log_level'):
            self.log_level = debug.LOG_LEVEL

    def enabled(self, context):
        """
        Disable rule.

        :param context:
        :type context:
        :return: True if rule is enabled, False if disabled
        :rtype: bool
        """
        return True

    def __lt__(self, other):
        return self.priority > other.priority

    def __repr__(self):
        defined = ""
        if self.defined_at:
            defined = f"@{self.defined_at}"
        return f"<{self.name if self.name else self.__class__.__name__}{defined}>"

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def __hash__(self):
        return hash(self.__class__)


class Rule(CustomRule):
    """
    Definition of a rule to apply
    """
    # pylint:disable=abstract-method
    consequence = None

    def then(self, matches, when_response, context):
        assert self.consequence
        if is_iterable(self.consequence):
            if not is_iterable(when_response):
                when_response = [when_response]
            iterator = iter(when_response)
            for cons in self.consequence:  #pylint: disable=not-an-iterable
                if inspect.isclass(cons):
                    cons = cons()
                cons.then(matches, next(iterator), context)
        else:
            cons = self.consequence
            if inspect.isclass(cons):
                cons = cons()  # pylint:disable=not-callable
            cons.then(matches, when_response, context)


class RemoveMatch(Consequence):  # pylint: disable=abstract-method
    """
    Remove matches returned by then
    """
    def then(self, matches, when_response, context):
        if is_iterable(when_response):
            ret = []
            when_response = list(when_response)
            for match in when_response:
                if match in matches:
                    matches.remove(match)
                    ret.append(match)
            return ret
        if when_response in matches:
            matches.remove(when_response)
            return when_response


class AppendMatch(Consequence):  # pylint: disable=abstract-method
    """
    Append matches returned by then
    """
    def __init__(self, match_name=None):
        self.match_name = match_name

    def then(self, matches, when_response, context):
        if is_iterable(when_response):
            ret = []
            when_response = list(when_response)
            for match in when_response:
                if match not in matches:
                    if self.match_name:
                        match.name = self.match_name
                    matches.append(match)
                    ret.append(match)
            return ret
        if self.match_name:
            when_response.name = self.match_name
        if when_response not in matches:
            matches.append(when_response)
            return when_response


class RenameMatch(Consequence):  # pylint: disable=abstract-method
    """
    Rename matches returned by then
    """
    def __init__(self, match_name):
        self.match_name = match_name
        self.remove = RemoveMatch()
        self.append = AppendMatch()

    def then(self, matches, when_response, context):
        removed = self.remove.then(matches, when_response, context)
        if is_iterable(removed):
            removed = list(removed)
            for match in removed:
                match.name = self.match_name
        elif removed:
            removed.name = self.match_name
        if removed:
            self.append.then(matches, removed, context)


class AppendTags(Consequence):  # pylint: disable=abstract-method
    """
    Add tags to returned matches
    """
    def __init__(self, tags):
        self.tags = tags
        self.remove = RemoveMatch()
        self.append = AppendMatch()

    def then(self, matches, when_response, context):
        removed = self.remove.then(matches, when_response, context)
        if is_iterable(removed):
            removed = list(removed)
            for match in removed:
                match.tags.extend(self.tags)
        elif removed:
            removed.tags.extend(self.tags)  # pylint: disable=no-member
        if removed:
            self.append.then(matches, removed, context)


class RemoveTags(Consequence):  # pylint: disable=abstract-method
    """
    Remove tags from returned matches
    """
    def __init__(self, tags):
        self.tags = tags
        self.remove = RemoveMatch()
        self.append = AppendMatch()

    def then(self, matches, when_response, context):
        removed = self.remove.then(matches, when_response, context)
        if is_iterable(removed):
            removed = list(removed)
            for match in removed:
                for tag in self.tags:
                    if tag in match.tags:
                        match.tags.remove(tag)
        elif removed:
            for tag in self.tags:
                if tag in removed.tags:  # pylint: disable=no-member
                    removed.tags.remove(tag)  # pylint: disable=no-member
        if removed:
            self.append.then(matches, removed, context)


class Rules(list):
    """
    list of rules ready to execute.
    """

    def __init__(self, *rules):
        super().__init__()
        self.load(*rules)

    def load(self, *rules):
        """
        Load rules from a Rule module, class or instance

        :param rules:
        :type rules:
        :return:
        :rtype:
        """
        for rule in rules:
            if inspect.ismodule(rule):
                self.load_module(rule)
            elif inspect.isclass(rule):
                self.load_class(rule)
            else:
                self.append(rule)

    def load_module(self, module):
        """
        Load a rules module

        :param module:
        :type module:
        :return:
        :rtype:
        """
        # pylint: disable=unused-variable
        for name, obj in inspect.getmembers(module,
                                            lambda member: hasattr(member, '__module__')
                                            and member.__module__ == module.__name__
                                            and inspect.isclass):
            self.load_class(obj)

    def load_class(self, class_):
        """
        Load a Rule class.

        :param class_:
        :type class_:
        :return:
        :rtype:
        """
        self.append(class_())

    def execute_all_rules(self, matches, context):
        """
        Execute all rules from this rules list. All when condition with same priority will be performed before
        calling then actions.

        :param matches:
        :type matches:
        :param context:
        :type context:
        :return:
        :rtype:
        """
        ret = []
        for priority, priority_rules in groupby(sorted(self), lambda rule: rule.priority):
            sorted_rules = toposort_rules(list(priority_rules))  # Group by dependency graph toposort
            for rules_group in sorted_rules:
                rules_group = list(sorted(rules_group, key=self.index))  # Sort rules group based on initial ordering.
                group_log_level = None
                for rule in rules_group:
                    if group_log_level is None or group_log_level < rule.log_level:
                        group_log_level = rule.log_level
                log(group_log_level, "%s independent rule(s) at priority %s.", len(rules_group), priority)
                for rule in rules_group:
                    when_response = execute_rule(rule, matches, context)
                    if when_response is not None:
                        ret.append((rule, when_response))

        return ret


def execute_rule(rule, matches, context):
    """
    Execute the given rule.
    :param rule:
    :type rule:
    :param matches:
    :type matches:
    :param context:
    :type context:
    :return:
    :rtype:
    """
    if rule.enabled(context):
        log(rule.log_level, "Checking rule condition: %s", rule)
        when_response = rule.when(matches, context)
        if when_response:
            log(rule.log_level, "Rule was triggered: %s", when_response)
            log(rule.log_level, "Running rule consequence: %s %s", rule, when_response)
            rule.then(matches, when_response, context)
            return when_response
    else:
        log(rule.log_level, "Rule is disabled: %s", rule)

def toposort_rules(rules):
    """
    Sort given rules using toposort with dependency parameter.
    :param rules:
    :type rules:
    :return:
    :rtype:
    """
    graph = {}
    class_dict = {}
    for rule in rules:
        if rule.__class__ in class_dict:
            raise ValueError(f"Duplicate class rules are not allowed: {rule.__class__}")
        class_dict[rule.__class__] = rule
    for rule in rules:
        if not is_iterable(rule.dependency) and rule.dependency:
            rule_dependencies = [rule.dependency]
        else:
            rule_dependencies = rule.dependency
        dependencies = set()
        if rule_dependencies:
            for dependency in rule_dependencies:
                if inspect.isclass(dependency):
                    dependency = class_dict.get(dependency)
                if dependency:
                    dependencies.add(dependency)
        graph[rule] = dependencies
    return toposort(graph)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Introspector tests
"""
# pylint: disable=pointless-statement,missing-docstring,protected-access,invalid-name,len-as-condition
from ..rebulk import Rebulk
from .. import introspector
from .default_rules_module import RuleAppend2, RuleAppend3


def test_string_introspector():
    rebulk = Rebulk().string('One', 'Two', 'Three', name='first').string('1', '2', '3', name='second')

    introspected = introspector.introspect(rebulk, None)

    assert len(introspected.patterns) == 2

    first_properties = introspected.patterns[0].properties
    assert len(first_properties) == 1
    first_properties['first'] == ['One', 'Two', 'Three']

    second_properties = introspected.patterns[1].properties
    assert len(second_properties) == 1
    second_properties['second'] == ['1', '2', '3']

    properties = introspected.properties
    assert len(properties) == 2
    assert properties['first'] == first_properties['first']
    assert properties['second'] == second_properties['second']


def test_string_properties():
    rebulk = Rebulk()\
        .string('One', 'Two', 'Three', name='first', properties={'custom': ['One']})\
        .string('1', '2', '3', name='second', properties={'custom': [1]})

    introspected = introspector.introspect(rebulk, None)

    assert len(introspected.patterns) == 2
    assert len(introspected.rules) == 2

    first_properties = introspected.patterns[0].properties
    assert len(first_properties) == 1
    first_properties['custom'] == ['One']

    second_properties = introspected.patterns[1].properties
    assert len(second_properties) == 1
    second_properties['custom'] == [1]

    properties = introspected.properties
    assert len(properties) == 1
    assert properties['custom'] == ['One', 1]


def test_various_pattern():
    rebulk = Rebulk()\
        .regex('One', 'Two', 'Three', name='first', value="string") \
        .string('1', '2', '3', name='second', value="digit") \
        .string('4', '5', '6', name='third') \
        .string('private', private=True) \
        .functional(lambda string: (0, 5), name='func', value='test') \
        .regex('One', 'Two', 'Three', name='regex_name') \
        .regex('(?P<one>One)(?P<two>Two)(?P<three>Three)') \
        .functional(lambda string: (6, 10), name='func2') \
        .string('7', name='third')

    introspected = introspector.introspect(rebulk, None)

    assert len(introspected.patterns) == 8
    assert len(introspected.rules) == 2

    first_properties = introspected.patterns[0].properties
    assert len(first_properties) == 1
    first_properties['first'] == ['string']

    second_properties = introspected.patterns[1].properties
    assert len(second_properties) == 1
    second_properties['second'] == ['digit']

    third_properties = introspected.patterns[2].properties
    assert len(third_properties) == 1
    third_properties['third'] == ['4', '5', '6']

    func_properties = introspected.patterns[3].properties
    assert len(func_properties) == 1
    func_properties['func'] == ['test']

    regex_name_properties = introspected.patterns[4].properties
    assert len(regex_name_properties) == 1
    regex_name_properties['regex_name'] == [None]

    regex_groups_properties = introspected.patterns[5].properties
    assert len(regex_groups_properties) == 3
    regex_groups_properties['one'] == [None]
    regex_groups_properties['two'] == [None]
    regex_groups_properties['three'] == [None]

    func2_properties = introspected.patterns[6].properties
    assert len(func2_properties) == 1
    func2_properties['func2'] == [None]

    append_third_properties = introspected.patterns[7].properties
    assert len(append_third_properties) == 1
    append_third_properties['third'] == [None]

    properties = introspected.properties
    assert len(properties) == 9
    assert properties['first'] == first_properties['first']
    assert properties['second'] == second_properties['second']
    assert properties['third'] == third_properties['third'] + append_third_properties['third']
    assert properties['func'] == func_properties['func']
    assert properties['regex_name'] == regex_name_properties['regex_name']
    assert properties['one'] == regex_groups_properties['one']
    assert properties['two'] == regex_groups_properties['two']
    assert properties['three'] == regex_groups_properties['three']
    assert properties['func2'] == func2_properties['func2']


def test_rule_properties():
    rebulk = Rebulk(default_rules=False).rules(RuleAppend2, RuleAppend3)

    introspected = introspector.introspect(rebulk, None)

    assert len(introspected.rules) == 2
    assert len(introspected.patterns) == 0

    rule_properties = introspected.rules[0].properties
    assert len(rule_properties) == 1
    assert rule_properties['renamed'] == [None]

    rule_properties = introspected.rules[1].properties
    assert len(rule_properties) == 1
    assert rule_properties['renamed'] == [None]

    properties = introspected.properties
    assert len(properties) == 1
    assert properties['renamed'] == [None]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Debug tools.

Can be configured by changing values of those variable.

DEBUG = False
Enable this variable to activate debug features (like defined_at parameters). It can slow down Rebulk

LOG_LEVEL = 0
Default log level of generated rebulk logs.
"""

import inspect
import logging
import os
from collections import namedtuple


DEBUG = False
LOG_LEVEL = logging.DEBUG


class Frame(namedtuple('Frame', ['lineno', 'package', 'name', 'filename'])):
    """
    Stack frame representation.
    """
    __slots__ = ()

    def __repr__(self):
        return f"{os.path.basename(self.filename)}#L{self.lineno}"


def defined_at():
    """
    Get definition location of a pattern or a match (outside of rebulk package).
    :return:
    :rtype:
    """
    if DEBUG:
        frame = inspect.currentframe()
        while frame:
            try:
                if frame.f_globals['__package__'] != __package__:
                    break
            except KeyError:  # pragma:no cover
                # If package is missing, consider we are in. Workaround for python 3.3.
                break
            frame = frame.f_back
        ret = Frame(frame.f_lineno,
                    frame.f_globals.get('__package__'),
                    frame.f_globals.get('__name__'),
                    frame.f_code.co_filename)
        del frame
        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Processor functions
"""
from logging import getLogger

from .utils import IdentitySet

from .rules import Rule, RemoveMatch

log = getLogger(__name__).log

DEFAULT = '__default__'

POST_PROCESS = -2048
PRE_PROCESS = 2048


def _default_conflict_solver(match, conflicting_match):
    """
    Default conflict solver for matches, shorter matches if they conflicts with longer ones

    :param conflicting_match:
    :type conflicting_match:
    :param match:
    :type match:
    :return:
    :rtype:
    """
    if len(conflicting_match.initiator) < len(match.initiator):
        return conflicting_match
    if len(match.initiator) < len(conflicting_match.initiator):
        return match
    return None


class ConflictSolver(Rule):
    """
    Remove conflicting matches.
    """
    priority = PRE_PROCESS

    consequence = RemoveMatch

    @property
    def default_conflict_solver(self):
        """
        Default conflict solver to use.
        """
        return _default_conflict_solver

    def when(self, matches, context):
        # pylint:disable=too-many-nested-blocks
        to_remove_matches = IdentitySet()

        public_matches = [match for match in matches if not match.private]
        public_matches.sort(key=len)

        for match in public_matches:
            conflicting_matches = matches.conflicting(match)

            if conflicting_matches:
                # keep the match only if it's the longest
                conflicting_matches = [conflicting_match for conflicting_match in conflicting_matches if
                                       not conflicting_match.private]
                conflicting_matches.sort(key=len)

                for conflicting_match in conflicting_matches:
                    conflict_solvers = [(self.default_conflict_solver, False)]

                    if match.conflict_solver:
                        conflict_solvers.append((match.conflict_solver, False))
                    if conflicting_match.conflict_solver:
                        conflict_solvers.append((conflicting_match.conflict_solver, True))

                    for conflict_solver, reverse in reversed(conflict_solvers):
                        if reverse:
                            to_remove = conflict_solver(conflicting_match, match)
                        else:
                            to_remove = conflict_solver(match, conflicting_match)
                        if to_remove == DEFAULT:
                            continue
                        if to_remove and to_remove not in to_remove_matches:
                            both_matches = [match, conflicting_match]
                            both_matches.remove(to_remove)
                            to_keep = both_matches[0]

                            if to_keep not in to_remove_matches:
                                log(self.log_level, "Conflicting match %s will be removed in favor of match %s",
                                    to_remove, to_keep)

                                to_remove_matches.add(to_remove)
                        break
        return to_remove_matches


class PrivateRemover(Rule):
    """
    Removes private matches rule.
    """
    priority = POST_PROCESS

    consequence = RemoveMatch

    def when(self, matches, context):
        return [match for match in matches if match.private]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2014 True Blade Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Original:
#   - https://bitbucket.org/ericvsmith/toposort (1.4)
# Modifications:
#   - port to pytest
# pylint: skip-file

import pytest
from ..toposort import toposort, toposort_flatten, CyclicDependency


class TestCase:
    def test_simple(self):
        results = list(toposort({2: set([11]), 9: set([11, 8]), 10: set([11, 3]), 11: set([7, 5]), 8: set([7, 3])}))
        expected = [set([3, 5, 7]), set([8, 11]), set([2, 9, 10])]
        assert results == expected

        # make sure self dependencies are ignored
        results = list(toposort({2: set([2, 11]), 9: set([11, 8]), 10: set([10, 11, 3]), 11: set([7, 5]), 8: set([7, 3])}))
        expected = [set([3, 5, 7]), set([8, 11]), set([2, 9, 10])]
        assert results == expected

        assert list(toposort({1: set()})) == [set([1])]
        assert list(toposort({1: set([1])})) == [set([1])]

    def test_no_dependencies(self):
        assert list(toposort({1: set([2]), 3: set([4]), 5: set([6])})) == [set([2, 4, 6]), set([1, 3, 5])]
        assert list(toposort({1: set(), 3: set(), 5: set()})) == [set([1, 3, 5])]

    def test_empty(self):
        assert list(toposort({})) == []

    def test_strings(self):
        results = list(toposort({'2': set(['11']), '9': set(['11', '8']), '10': set(['11', '3']), '11': set(['7', '5']), '8': set(['7', '3'])}))
        expected = [set(['3', '5', '7']), set(['8', '11']), set(['2', '9', '10'])]
        assert results == expected

    def test_objects(self):
        o2 = object()
        o3 = object()
        o5 = object()
        o7 = object()
        o8 = object()
        o9 = object()
        o10 = object()
        o11 = object()
        results = list(toposort({o2: set([o11]), o9: set([o11, o8]), o10: set([o11, o3]), o11: set([o7, o5]), o8: set([o7, o3, o8])}))
        expected = [set([o3, o5, o7]), set([o8, o11]), set([o2, o9, o10])]
        assert results == expected

    def test_cycle(self):
        # a simple, 2 element cycle
        with pytest.raises(CyclicDependency):
            list(toposort({1: set([2]), 2: set([1])}))

        # an indirect cycle
        with pytest.raises(CyclicDependency):
            list(toposort({1: set([2]), 2: set([3]), 3: set([1])}))

    def test_input_not_modified(self):
        data = {2: set([11]),
                9: set([11, 8]),
                10: set([11, 3]),
                11: set([7, 5]),
                8: set([7, 3, 8]),  # includes something self-referential
                }
        orig = data.copy()
        results = list(toposort(data))
        assert data == orig

    def test_input_not_modified_when_cycle_error(self):
        data = {1: set([2]),
                2: set([1]),
                3: set([4]),
                }
        orig = data.copy()
        with pytest.raises(CyclicDependency):
            list(toposort(data))
        assert data == orig


class TestCaseAll:
    def test_sort_flatten(self):
        data = {2: set([11]),
                9: set([11, 8]),
                10: set([11, 3]),
                11: set([7, 5]),
                8: set([7, 3, 8]),  # includes something self-referential
                }
        expected = [set([3, 5, 7]), set([8, 11]), set([2, 9, 10])]
        assert list(toposort(data)) == expected

        # now check the sorted results
        results = []
        for item in expected:
            results.extend(sorted(item))
        assert toposort_flatten(data) == results

        # and the unsorted results. break the results up into groups to compare them
        actual = toposort_flatten(data, False)
        results = [set([i for i in actual[0:3]]), set([i for i in actual[3:5]]), set([i for i in actual[5:8]])]
        assert results == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Define simple search patterns in bulk to perform advanced matching on any string.
"""
#  pylint:disable=import-self
from .rebulk import Rebulk
from .rules import Rule, CustomRule, AppendMatch, RemoveMatch, RenameMatch, AppendTags, RemoveTags
from .processors import ConflictSolver, PrivateRemover, POST_PROCESS, PRE_PROCESS
from .pattern import REGEX_ENABLED



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, invalid-name, no-member, len-as-condition
import pytest
from rebulk.test.default_rules_module import RuleRemove0, RuleAppend0, RuleRename0, RuleAppend1, RuleRemove1, \
    RuleRename1, RuleAppend2, RuleRename2, RuleAppend3, RuleRename3, RuleAppendTags0, RuleRemoveTags0, \
    RuleAppendTags1, RuleRemoveTags1

from ..rules import Rules
from ..match import Matches, Match

from .rules_module import Rule1, Rule2, Rule3, Rule0, Rule1Disabled
from . import rules_module as rm


def test_rule_priority():
    matches = Matches([Match(1, 2)])

    rules = Rules(Rule1, Rule2())

    rules.execute_all_rules(matches, {})
    assert len(matches) == 0
    matches = Matches([Match(1, 2)])

    rules = Rules(Rule1(), Rule0)

    rules.execute_all_rules(matches, {})
    assert len(matches) == 1
    assert matches[0] == Match(3, 4)


def test_rules_duplicates():
    matches = Matches([Match(1, 2)])

    rules = Rules(Rule1, Rule1)

    with pytest.raises(ValueError):
        rules.execute_all_rules(matches, {})


def test_rule_disabled():
    matches = Matches([Match(1, 2)])

    rules = Rules(Rule1Disabled(), Rule2())

    rules.execute_all_rules(matches, {})
    assert len(matches) == 2
    assert matches[0] == Match(1, 2)
    assert matches[1] == Match(3, 4)


def test_rule_when():
    matches = Matches([Match(1, 2)])

    rules = Rules(Rule3())

    rules.execute_all_rules(matches, {'when': False})
    assert len(matches) == 1
    assert matches[0] == Match(1, 2)

    matches = Matches([Match(1, 2)])

    rules.execute_all_rules(matches, {'when': True})
    assert len(matches) == 2
    assert matches[0] == Match(1, 2)
    assert matches[1] == Match(3, 4)


class TestDefaultRules:
    def test_remove(self):
        rules = Rules(RuleRemove0)

        matches = Matches([Match(1, 2)])
        rules.execute_all_rules(matches, {})

        assert len(matches) == 0

        rules = Rules(RuleRemove1)

        matches = Matches([Match(1, 2)])
        rules.execute_all_rules(matches, {})

        assert len(matches) == 0

    def test_append(self):
        rules = Rules(RuleAppend0)

        matches = Matches([Match(1, 2)])
        rules.execute_all_rules(matches, {})

        assert len(matches) == 2

        rules = Rules(RuleAppend1)

        matches = Matches([Match(1, 2)])
        rules.execute_all_rules(matches, {})

        assert len(matches) == 2

        rules = Rules(RuleAppend2)

        matches = Matches([Match(1, 2)])
        rules.execute_all_rules(matches, {})

        assert len(matches) == 2
        assert len(matches.named('renamed')) == 1

        rules = Rules(RuleAppend3)

        matches = Matches([Match(1, 2)])
        rules.execute_all_rules(matches, {})

        assert len(matches) == 2
        assert len(matches.named('renamed')) == 1

    def test_rename(self):
        rules = Rules(RuleRename0)

        matches = Matches([Match(1, 2, name='original')])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('original')) == 1
        assert len(matches.named('renamed')) == 0

        rules = Rules(RuleRename1)

        matches = Matches([Match(5, 10, name='original')])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('original')) == 0
        assert len(matches.named('renamed')) == 1

        rules = Rules(RuleRename2)

        matches = Matches([Match(5, 10, name='original')])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('original')) == 0
        assert len(matches.named('renamed')) == 1

        rules = Rules(RuleRename3)

        matches = Matches([Match(5, 10, name='original')])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('original')) == 0
        assert len(matches.named('renamed')) == 1

    def test_append_tags(self):
        rules = Rules(RuleAppendTags0)

        matches = Matches([Match(1, 2, name='tags', tags=['other'])])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('tags')) == 1
        assert matches.named('tags', index=0).tags == ['other', 'new-tag']

        rules = Rules(RuleAppendTags1)

        matches = Matches([Match(1, 2, name='tags', tags=['other'])])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('tags')) == 1
        assert matches.named('tags', index=0).tags == ['other', 'new-tag']

    def test_remove_tags(self):
        rules = Rules(RuleRemoveTags0)

        matches = Matches([Match(1, 2, name='tags', tags=['other', 'new-tag'])])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('tags')) == 1
        assert matches.named('tags', index=0).tags == ['other']

        rules = Rules(RuleRemoveTags1)

        matches = Matches([Match(1, 2, name='tags', tags=['other', 'new-tag'])])
        rules.execute_all_rules(matches, {})

        assert len(matches.named('tags')) == 1
        assert matches.named('tags', index=0).tags == ['other']


def test_rule_module():
    rules = Rules(rm)

    matches = Matches([Match(1, 2)])
    rules.execute_all_rules(matches, {})

    assert len(matches) == 1


def test_rule_repr():
    assert str(Rule0()) == "<Rule0>"
    assert str(Rule1()) == "<Rule1>"
    assert str(Rule2()) == "<Rule2>"
    assert str(Rule1Disabled()) == "<Disabled Rule1>"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Uniform re module
"""
# pylint: disable-all
import os
import logging

log = logging.getLogger(__name__).log

REGEX_ENABLED = False
if os.environ.get('REBULK_REGEX_ENABLED') in ["1", "true", "True", "Y"]:
    try:
        import regex as re
        REGEX_ENABLED = True
    except ImportError:
        log.warning('regex module is not available. Unset REBULK_REGEX_ENABLED environment variable, or install regex module to enabled it.')
        import re
else:
    import re



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, unneeded-not, len-as-condition

import pytest

from ..match import Match, Matches
from ..pattern import StringPattern, RePattern
from ..formatters import formatters


class TestMatchClass:
    def test_repr(self):
        match1 = Match(1, 3, value="es")

        assert repr(match1) == '<es:(1, 3)>'

        match2 = Match(0, 4, value="test", private=True, name="abc", tags=['one', 'two'])

        assert repr(match2) == '<test:(0, 4)+private+name=abc+tags=[\'one\', \'two\']>'

    def test_names(self):
        parent = Match(0, 10, name="test")
        parent.children.append(Match(0, 10, name="child1", parent=parent))
        parent.children.append(Match(0, 10, name="child2", parent=parent))

        assert set(parent.names) == set(["child1", "child2"])

    def test_equality(self):
        match1 = Match(1, 3, value="es")
        match2 = Match(1, 3, value="es")

        other = object()

        assert hash(match1) == hash(match2)
        assert hash(match1) != hash(other)

        assert match1 == match2
        assert not match1 == other

    def test_inequality(self):
        match1 = Match(0, 2, value="te")
        match2 = Match(2, 4, value="st")
        match3 = Match(0, 2, value="other")

        other = object()

        assert hash(match1) != hash(match2)
        assert hash(match1) != hash(match3)

        assert match1 != other
        assert match1 != match2
        assert match1 != match3

    def test_length(self):
        match1 = Match(0, 4, value="test")
        match2 = Match(0, 2, value="spanIsUsed")

        assert len(match1) == 4
        assert len(match2) == 2

    def test_compare(self):
        match1 = Match(0, 2, value="te")
        match2 = Match(2, 4, value="st")

        other = object()

        assert match1 < match2
        assert match1 <= match2

        assert match2 > match1
        assert match2 >= match1

        with pytest.raises(TypeError):
            match1 < other

        with pytest.raises(TypeError):
            match1 <= other

        with pytest.raises(TypeError):
            match1 > other

        with pytest.raises(TypeError):
            match1 >= other


    def test_value(self):
        match1 = Match(1, 3)
        match1.value = "test"

        assert match1.value == "test"


class TestMatchesClass:
    match1 = Match(0, 2, value="te", name="start")
    match2 = Match(2, 3, value="s", tags="tag1")
    match3 = Match(3, 4, value="t", tags=["tag1", "tag2"])
    match4 = Match(2, 4, value="st", name="end")

    def test_tag(self):
        matches = Matches()
        matches.append(self.match1)
        matches.append(self.match2)
        matches.append(self.match3)
        matches.append(self.match4)

        assert "start" in matches.names
        assert "end" in matches.names

        assert "tag1" in matches.tags
        assert "tag2" in matches.tags

        assert self.match3.tagged("tag1")
        assert not self.match3.tagged("start")

        tag1 = matches.tagged("tag1")
        assert len(tag1) == 2
        assert tag1[0] == self.match2
        assert tag1[1] == self.match3

        tag2 = matches.tagged("tag2")
        assert len(tag2) == 1
        assert tag2[0] == self.match3

        start = matches.named("start")
        assert len(start) == 1
        assert start[0] == self.match1

        end = matches.named("end")
        assert len(end) == 1
        assert end[0] == self.match4

    def test_base(self):
        matches = Matches()
        matches.append(self.match1)

        assert len(matches) == 1
        assert repr(matches) == repr([self.match1])
        assert list(matches.starting(0)) == [self.match1]
        assert list(matches.ending(2)) == [self.match1]

        matches.append(self.match2)
        matches.append(self.match3)
        matches.append(self.match4)

        assert len(matches) == 4
        assert list(matches.starting(2)) == [self.match2, self.match4]
        assert list(matches.starting(3)) == [self.match3]
        assert list(matches.ending(3)) == [self.match2]
        assert list(matches.ending(4)) == [self.match3, self.match4]
        assert list(matches.range()) == [self.match1, self.match2, self.match4, self.match3]
        assert list(matches.range(0)) == [self.match1, self.match2, self.match4, self.match3]
        assert list(matches.range(0, 3)) == [self.match1, self.match2, self.match4]
        assert list(matches.range(2, 3)) == [self.match2, self.match4]
        assert list(matches.range(3, 4)) == [self.match4, self.match3]

        matches.remove(self.match1)
        assert len(matches) == 3
        assert len(matches.starting(0)) == 0
        assert len(matches.ending(2)) == 0

        matches.clear()

        assert len(matches) == 0
        assert len(matches.starting(0)) == 0
        assert len(matches.starting(2)) == 0
        assert len(matches.starting(3)) == 0
        assert len(matches.ending(2)) == 0
        assert len(matches.ending(3)) == 0
        assert len(matches.ending(4)) == 0

    def test_get_slices(self):
        matches = Matches()
        matches.append(self.match1)
        matches.append(self.match2)
        matches.append(self.match3)
        matches.append(self.match4)

        slice_matches = matches[1:3]

        assert isinstance(slice_matches, Matches)

        assert len(slice_matches) == 2
        assert slice_matches[0] == self.match2
        assert slice_matches[1] == self.match3

    def test_remove_slices(self):
        matches = Matches()
        matches.append(self.match1)
        matches.append(self.match2)
        matches.append(self.match3)
        matches.append(self.match4)

        del matches[1:3]

        assert len(matches) == 2
        assert matches[0] == self.match1
        assert matches[1] == self.match4

    def test_set_slices(self):
        matches = Matches()
        matches.append(self.match1)
        matches.append(self.match2)
        matches.append(self.match3)
        matches.append(self.match4)

        matches[1:3] = self.match1, self.match4

        assert len(matches) == 4
        assert matches[0] == self.match1
        assert matches[1] == self.match1
        assert matches[2] == self.match4
        assert matches[3] == self.match4

    def test_set_index(self):
        matches = Matches()
        matches.append(self.match1)
        matches.append(self.match2)
        matches.append(self.match3)

        matches[1] = self.match4

        assert len(matches) == 3
        assert matches[0] == self.match1
        assert matches[1] == self.match4
        assert matches[2] == self.match3

    def test_constructor(self):
        matches = Matches([self.match1, self.match2, self.match3, self.match4])

        assert len(matches) == 4
        assert list(matches.starting(0)) == [self.match1]
        assert list(matches.ending(2)) == [self.match1]
        assert list(matches.starting(2)) == [self.match2, self.match4]
        assert list(matches.starting(3)) == [self.match3]
        assert list(matches.ending(3)) == [self.match2]
        assert list(matches.ending(4)) == [self.match3, self.match4]

    def test_constructor_kwargs(self):
        matches = Matches([self.match1, self.match2, self.match3, self.match4], input_string="test")

        assert len(matches) == 4
        assert matches.input_string == "test"
        assert list(matches.starting(0)) == [self.match1]
        assert list(matches.ending(2)) == [self.match1]
        assert list(matches.starting(2)) == [self.match2, self.match4]
        assert list(matches.starting(3)) == [self.match3]
        assert list(matches.ending(3)) == [self.match2]
        assert list(matches.ending(4)) == [self.match3, self.match4]

    def test_crop(self):
        input_string = "abcdefghijklmnopqrstuvwxyz"

        match1 = Match(1, 10, input_string=input_string)
        match2 = Match(0, 2, input_string=input_string)
        match3 = Match(8, 15, input_string=input_string)

        ret = match1.crop([match2, match3.span])

        assert len(ret) == 1

        assert ret[0].span == (2, 8)
        assert ret[0].value == "cdefgh"

        ret = match1.crop((1, 10))
        assert len(ret) == 0

        ret = match1.crop((1, 3))
        assert len(ret) == 1
        assert ret[0].span == (3, 10)

        ret = match1.crop((7, 10))
        assert len(ret) == 1
        assert ret[0].span == (1, 7)

        ret = match1.crop((0, 12))
        assert len(ret) == 0

        ret = match1.crop((4, 6))
        assert len(ret) == 2

        assert ret[0].span == (1, 4)
        assert ret[1].span == (6, 10)

        ret = match1.crop([(3, 5), (7, 9)])
        assert len(ret) == 3

        assert ret[0].span == (1, 3)
        assert ret[1].span == (5, 7)
        assert ret[2].span == (9, 10)

    def test_split(self):
        input_string = "123 +word1  -  word2  + word3  456"
        match = Match(3, len(input_string) - 3, input_string=input_string)
        splitted = match.split(" -+")

        assert len(splitted) == 3
        assert [split.value for split in splitted] == ["word1", "word2", "word3"]


class TestMaches:
    def test_names(self):
        input_string = "One Two Three"

        matches = Matches()

        matches.extend(StringPattern("One", name="1-str", tags=["One", "str"]).matches(input_string))
        matches.extend(RePattern("One", name="1-re", tags=["One", "re"]).matches(input_string))
        matches.extend(StringPattern("Two", name="2-str", tags=["Two", "str"]).matches(input_string))
        matches.extend(RePattern("Two", name="2-re", tags=["Two", "re"]).matches(input_string))
        matches.extend(StringPattern("Three", name="3-str", tags=["Three", "str"]).matches(input_string))
        matches.extend(RePattern("Three", name="3-re", tags=["Three", "re"]).matches(input_string))

        assert set(matches.names) == set(["1-str", "1-re", "2-str", "2-re", "3-str", "3-re"])

    def test_filters(self):
        input_string = "One Two Three"

        matches = Matches()

        matches.extend(StringPattern("One", name="1-str", tags=["One", "str"]).matches(input_string))
        matches.extend(RePattern("One", name="1-re", tags=["One", "re"]).matches(input_string))
        matches.extend(StringPattern("Two", name="2-str", tags=["Two", "str"]).matches(input_string))
        matches.extend(RePattern("Two", name="2-re", tags=["Two", "re"]).matches(input_string))
        matches.extend(StringPattern("Three", name="3-str", tags=["Three", "str"]).matches(input_string))
        matches.extend(RePattern("Three", name="3-re", tags=["Three", "re"]).matches(input_string))

        selection = matches.starting(0)
        assert len(selection) == 2

        selection = matches.starting(0, lambda m: "str" in m.tags)
        assert len(selection) == 1
        assert selection[0].pattern.name == "1-str"

        selection = matches.ending(7, predicate=lambda m: "str" in m.tags)
        assert len(selection) == 1
        assert selection[0].pattern.name == "2-str"

        selection = matches.previous(matches.named("2-str")[0])
        assert len(selection) == 2
        assert selection[0].pattern.name == "1-str"
        assert selection[1].pattern.name == "1-re"

        selection = matches.previous(matches.named("2-str", 0), lambda m: "str" in m.tags)
        assert len(selection) == 1
        assert selection[0].pattern.name == "1-str"

        selection = matches.next(matches.named("2-str", 0))
        assert len(selection) == 2
        assert selection[0].pattern.name == "3-str"
        assert selection[1].pattern.name == "3-re"

        selection = matches.next(matches.named("2-str", 0), index=0, predicate=lambda m: "re" in m.tags)
        assert selection is not None
        assert selection.pattern.name == "3-re"

        selection = matches.next(matches.named("2-str", index=0), lambda m: "re" in m.tags)
        assert len(selection) == 1
        assert selection[0].pattern.name == "3-re"

        selection = matches.named("2-str", lambda m: "re" in m.tags)
        assert len(selection) == 0

        selection = matches.named("2-re", lambda m: "re" in m.tags, 0)
        assert selection is not None
        assert selection.name == "2-re"  # pylint:disable=no-member

        selection = matches.named("2-re", lambda m: "re" in m.tags)
        assert len(selection) == 1
        assert selection[0].name == "2-re"

        selection = matches.named("2-re", lambda m: "re" in m.tags, index=1000)
        assert selection is None

    def test_raw(self):
        input_string = "0123456789"

        match = Match(0, 10, input_string=input_string, formatter=lambda s: s*2)

        assert match.value == match.raw * 2
        assert match.raw == input_string

        match.raw_end = 9
        match.raw_start = 1

        assert match.value == match.raw * 2
        assert match.raw == input_string[1:9]

        match.raw_end = None
        match.raw_start = None

        assert match.value == match.raw * 2
        assert match.raw == input_string


    def test_formatter_chain(self):
        input_string = "100"

        match = Match(0, 3, input_string=input_string, formatter=formatters(int, lambda s: s*2, lambda  s: s+10))

        assert match.raw == input_string
        assert match.value == 100 * 2 + 10


    def test_to_dict(self):
        input_string = "One Two Two Three"

        matches = Matches()

        matches.extend(StringPattern("One", name="1", tags=["One", "str"]).matches(input_string))
        matches.extend(RePattern("One", name="1", tags=["One", "re"]).matches(input_string))
        matches.extend(StringPattern("Two", name="2", tags=["Two", "str"]).matches(input_string))
        matches.extend(RePattern("Two", name="2", tags=["Two", "re"]).matches(input_string))
        matches.extend(RePattern("Two", name="2", tags=["Two", "reBis"]).matches(input_string))
        matches.extend(StringPattern("Three", name="3", tags=["Three", "str"]).matches(input_string))
        matches.extend(RePattern("Three", name="3bis", tags=["Three", "re"]).matches(input_string))
        matches.extend(RePattern(r"(\w+)", name="words").matches(input_string))

        kvalues = matches.to_dict(first_value=True)
        assert kvalues == {"1": "One",
                           "2": "Two",
                           "3": "Three",
                           "3bis": "Three",
                           "words": "One"}
        assert kvalues.values_list["words"] == ["One", "Two", "Three"]

        kvalues = matches.to_dict(enforce_list=True)
        assert kvalues["words"] == ["One", "Two", "Three"]

        kvalues = matches.to_dict(details=True)
        assert kvalues["1"].value == "One"

        assert len(kvalues["2"]) == 2
        assert kvalues["2"][0].value == "Two"
        assert kvalues["2"][1].value == "Two"

        assert kvalues["3"].value == "Three"
        assert kvalues["3bis"].value == "Three"

        assert len(kvalues["words"]) == 4
        assert kvalues["words"][0].value == "One"
        assert kvalues["words"][1].value == "Two"
        assert kvalues["words"][2].value == "Two"
        assert kvalues["words"][3].value == "Three"

        kvalues = matches.to_dict(details=True)
        assert kvalues["1"].value == "One"

        assert len(kvalues.values_list["2"]) == 2
        assert kvalues.values_list["2"][0].value == "Two"
        assert kvalues.values_list["2"][1].value == "Two"

        assert kvalues["3"].value == "Three"
        assert kvalues["3bis"].value == "Three"

        assert len(kvalues.values_list["words"]) == 4
        assert kvalues.values_list["words"][0].value == "One"
        assert kvalues.values_list["words"][1].value == "Two"
        assert kvalues.values_list["words"][2].value == "Two"
        assert kvalues.values_list["words"][3].value == "Three"

    def test_chains(self):
        input_string = "wordX 10 20 30 40 wordA, wordB, wordC 70 80 wordX"

        matches = Matches(input_string=input_string)

        matches.extend(RePattern(r"\d+", name="digit").matches(input_string))
        matches.extend(RePattern("[a-zA-Z]+", name="word").matches(input_string))

        assert len(matches) == 11

        a_start = input_string.find('wordA')

        b_start = input_string.find('wordB')
        b_end = b_start + len('wordB')

        c_start = input_string.find('wordC')
        c_end = c_start + len('wordC')

        chain_before = matches.chain_before(b_start, " ,", predicate=lambda match: match.name == "word")
        assert len(chain_before) == 1
        assert chain_before[0].value == 'wordA'

        chain_before = matches.chain_before(Match(b_start, b_start), " ,", predicate=lambda match: match.name == "word")
        assert len(chain_before) == 1
        assert chain_before[0].value == 'wordA'

        chain_before = matches.chain_before(b_start, " ,", predicate=lambda match: match.name == "digit")
        assert len(chain_before) == 0

        chain_before = matches.chain_before(a_start, " ,", predicate=lambda match: match.name == "digit")
        assert len(chain_before) == 4
        assert [match.value for match in chain_before] == ["40", "30", "20", "10"]

        chain_after = matches.chain_after(b_end, " ,", predicate=lambda match: match.name == "word")
        assert len(chain_after) == 1
        assert chain_after[0].value == 'wordC'

        chain_after = matches.chain_after(Match(b_end, b_end), " ,", predicate=lambda match: match.name == "word")
        assert len(chain_after) == 1
        assert chain_after[0].value == 'wordC'

        chain_after = matches.chain_after(b_end, " ,", predicate=lambda match: match.name == "digit")
        assert len(chain_after) == 0

        chain_after = matches.chain_after(c_end, " ,", predicate=lambda match: match.name == "digit")
        assert len(chain_after) == 2
        assert [match.value for match in chain_after] == ["70", "80"]

        chain_after = matches.chain_after(c_end, " ,", end=10000, predicate=lambda match: match.name == "digit")
        assert len(chain_after) == 2
        assert [match.value for match in chain_after] == ["70", "80"]

    def test_holes(self):
        input_string = '1'*10+'2'*10+'3'*10+'4'*10+'5'*10+'6'*10+'7'*10

        hole1 = Match(0, 10, input_string=input_string)
        hole2 = Match(20, 30, input_string=input_string)
        hole3 = Match(30, 40, input_string=input_string)
        hole4 = Match(60, 70, input_string=input_string)

        matches = Matches([hole1, hole2], input_string=input_string)
        matches.append(hole3)
        matches.append(hole4)

        holes = list(matches.holes())
        assert len(holes) == 2
        assert holes[0].span == (10, 20)
        assert holes[0].value == '2'*10
        assert holes[1].span == (40, 60)
        assert holes[1].value == '5' * 10 + '6' * 10

        holes = list(matches.holes(5, 15))
        assert len(holes) == 1
        assert holes[0].span == (10, 15)
        assert holes[0].value == '2'*5

        holes = list(matches.holes(5, 15, formatter=lambda value: "formatted"))
        assert len(holes) == 1
        assert holes[0].span == (10, 15)
        assert holes[0].value == "formatted"

        holes = list(matches.holes(5, 15, predicate=lambda hole: False))
        assert len(holes) == 0

    def test_holes_empty(self):
        input_string = "Test hole on empty matches"
        matches = Matches(input_string=input_string)
        holes = matches.holes()
        assert len(holes) == 1
        assert holes[0].value == input_string

    def test_holes_seps(self):
        input_string = "Test hole - with many separators + included"
        match = StringPattern("many").matches(input_string)

        matches = Matches(match, input_string)
        holes = matches.holes()

        assert len(holes) == 2

        holes = matches.holes(seps="-+")

        assert len(holes) == 4
        assert [hole.value for hole in holes] == ["Test hole ", " with ", " separators ", " included"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, invalid-name,len-as-condition

from functools import partial

from rebulk.pattern import StringPattern

from ..validators import chars_before, chars_after, chars_surround, validators

chars = ' _.'
left = partial(chars_before, chars)
right = partial(chars_after, chars)
surrounding = partial(chars_surround, chars)


def test_left_chars():
    matches = list(StringPattern("word", validator=left).matches("xxxwordxxx"))
    assert len(matches) == 0

    matches = list(StringPattern("word", validator=left).matches("xxx_wordxxx"))
    assert len(matches) == 1

    matches = list(StringPattern("word", validator=left).matches("wordxxx"))
    assert len(matches) == 1


def test_right_chars():
    matches = list(StringPattern("word", validator=right).matches("xxxwordxxx"))
    assert len(matches) == 0

    matches = list(StringPattern("word", validator=right).matches("xxxword.xxx"))
    assert len(matches) == 1

    matches = list(StringPattern("word", validator=right).matches("xxxword"))
    assert len(matches) == 1


def test_surrounding_chars():
    matches = list(StringPattern("word", validator=surrounding).matches("xxxword xxx"))
    assert len(matches) == 0

    matches = list(StringPattern("word", validator=surrounding).matches("xxx.wordxxx"))
    assert len(matches) == 0

    matches = list(StringPattern("word", validator=surrounding).matches("xxx word_xxx"))
    assert len(matches) == 1

    matches = list(StringPattern("word", validator=surrounding).matches("word"))
    assert len(matches) == 1


def test_chain():
    matches = list(StringPattern("word", validator=validators(left, right)).matches("xxxword xxx"))
    assert len(matches) == 0

    matches = list(StringPattern("word", validator=validators(left, right)).matches("xxx.wordxxx"))
    assert len(matches) == 0

    matches = list(StringPattern("word", validator=validators(left, right)).matches("xxx word_xxx"))
    assert len(matches) == 1

    matches = list(StringPattern("word", validator=validators(left, right)).matches("word"))
    assert len(matches) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Chain patterns and handle repetiting capture group
"""
import itertools

from .builder import Builder
from .loose import call
from .match import Match, Matches
from .pattern import Pattern, filter_match_kwargs, BasePattern
from .remodule import re


class _InvalidChainException(Exception):
    """
    Internal exception raised when a chain is not valid
    """


class Chain(Pattern, Builder):
    """
    Definition of a pattern chain to search for.
    """

    def __init__(self, parent, chain_breaker=None, **kwargs):  # pylint: disable=super-init-not-called
        Builder.__init__(self)
        call(Pattern.__init__, self, **kwargs)
        self._kwargs = kwargs
        self._match_kwargs = filter_match_kwargs(kwargs)
        if callable(chain_breaker):
            self.chain_breaker = chain_breaker
        else:
            self.chain_breaker = None
        self.parent = parent
        self.parts = []

    def pattern(self, *pattern):
        """

        :param pattern:
        :return:
        """
        if not pattern:
            raise ValueError("One pattern should be given to the chain")
        if len(pattern) > 1:
            raise ValueError("Only one pattern can be given to the chain")
        part = ChainPart(self, pattern[0])
        self.parts.append(part)
        return part

    def close(self):
        """
        Deeply close the chain
        :return: Rebulk instance
        """
        parent = self.parent
        while isinstance(parent, Chain):
            parent = parent.parent
        return parent

    def _match(self, pattern, input_string, context=None):
        # pylint: disable=too-many-locals,too-many-nested-blocks
        chain_matches = []
        chain_input_string = input_string
        offset = 0
        while offset < len(input_string):
            chain_found = False
            current_chain_matches = []
            valid_chain = True
            for chain_part in self.parts:
                try:
                    chain_part_matches, raw_chain_part_matches = chain_part.matches(chain_input_string,
                                                                                    context,
                                                                                    with_raw_matches=True)

                    chain_found, chain_input_string, offset = \
                        self._to_next_chain_part(chain_part, chain_part_matches, raw_chain_part_matches, chain_found,
                                                 input_string, chain_input_string, offset, current_chain_matches)
                except _InvalidChainException:
                    valid_chain = False
                    if current_chain_matches:
                        offset = current_chain_matches[0].raw_end
                    break
            if not chain_found:
                break
            if current_chain_matches and valid_chain:
                match = self._build_chain_match(current_chain_matches, input_string)
                chain_matches.append(match)

        return chain_matches

    def _to_next_chain_part(self, chain_part, chain_part_matches, raw_chain_part_matches, chain_found,
                            input_string, chain_input_string, offset, current_chain_matches):
        Chain._fix_matches_offset(chain_part_matches, input_string, offset)
        Chain._fix_matches_offset(raw_chain_part_matches, input_string, offset)

        if raw_chain_part_matches:
            grouped_matches_dict = self._group_by_match_index(chain_part_matches)
            grouped_raw_matches_dict = self._group_by_match_index(raw_chain_part_matches)

            for match_index, grouped_raw_matches in grouped_raw_matches_dict.items():
                chain_found = True
                offset = grouped_raw_matches[-1].raw_end
                chain_input_string = input_string[offset:]

                if not chain_part.is_hidden:
                    grouped_matches = grouped_matches_dict.get(match_index, [])
                    if self._chain_breaker_eval(current_chain_matches + grouped_matches):
                        current_chain_matches.extend(grouped_matches)
        return chain_found, chain_input_string, offset

    def _process_match(self, match, match_index, child=False):
        """
        Handle a match
        :param match:
        :type match:
        :param match_index:
        :type match_index:
        :param child:
        :type child:
        :return:
        :rtype:
        """
        # pylint: disable=too-many-locals
        ret = super()._process_match(match, match_index, child=child)
        if ret:
            return True

        if match.children:
            last_pattern = match.children[-1].pattern
            last_pattern_groups = self._group_by_match_index(
                [child_ for child_ in match.children if child_.pattern == last_pattern]
            )

            if last_pattern_groups:
                original_children = Matches(match.children)
                original_end = match.end

                for index in reversed(list(last_pattern_groups)):
                    last_matches = last_pattern_groups[index]
                    for last_match in last_matches:
                        match.children.remove(last_match)
                    match.end = match.children[-1].end if match.children else match.start
                    ret = super()._process_match(match, match_index, child=child)
                    if ret:
                        return True

                match.children = original_children
                match.end = original_end

        return False

    def _build_chain_match(self, current_chain_matches, input_string):
        start = None
        end = None
        for match in current_chain_matches:
            if start is None or start > match.start:
                start = match.start
            if end is None or end < match.end:
                end = match.end
        match = call(Match, start, end, pattern=self, input_string=input_string, **self._match_kwargs)
        for chain_match in current_chain_matches:
            if chain_match.children:
                for child in chain_match.children:
                    match.children.append(child)
            if chain_match not in match.children:
                match.children.append(chain_match)
                chain_match.parent = match
        return match

    def _chain_breaker_eval(self, matches):
        return not self.chain_breaker or not self.chain_breaker(Matches(matches))

    @staticmethod
    def _fix_matches_offset(chain_part_matches, input_string, offset):
        for chain_part_match in chain_part_matches:
            if chain_part_match.input_string != input_string:
                chain_part_match.input_string = input_string
                chain_part_match.end += offset
                chain_part_match.start += offset
            if chain_part_match.children:
                Chain._fix_matches_offset(chain_part_match.children, input_string, offset)

    @staticmethod
    def _group_by_match_index(matches):
        grouped_matches_dict = {}
        for match_index, match in itertools.groupby(matches, lambda m: m.match_index):
            grouped_matches_dict[match_index] = list(match)
        return grouped_matches_dict

    @property
    def match_options(self):
        return {}

    @property
    def patterns(self):
        return [self]

    def __repr__(self):
        defined = ""
        if self.defined_at:
            defined = f"@{self.defined_at}"
        return f"<{self.__class__.__name__}{defined}:{self.parts}>"


class ChainPart(BasePattern):
    """
    Part of a pattern chain.
    """

    def __init__(self, chain, pattern):
        self._chain = chain
        self.pattern = pattern
        self.repeater_start = 1
        self.repeater_end = 1
        self._hidden = False

    @property
    def _is_chain_start(self):
        return self._chain.parts[0] == self

    def matches(self, input_string, context=None, with_raw_matches=False):
        matches, raw_matches = self.pattern.matches(input_string, context=context, with_raw_matches=True)

        matches = self._truncate_repeater(matches, input_string)
        raw_matches = self._truncate_repeater(raw_matches, input_string)

        self._validate_repeater(raw_matches)

        if with_raw_matches:
            return matches, raw_matches

        return matches

    def _truncate_repeater(self, matches, input_string):
        if not matches:
            return matches

        if not self._is_chain_start:
            separator = input_string[0:matches[0].initiator.raw_start]
            if separator:
                return []

        j = 1
        for i in range(0, len(matches) - 1):
            separator = input_string[matches[i].initiator.raw_end:
                                     matches[i + 1].initiator.raw_start]
            if separator:
                break
            j += 1
        truncated = matches[:j]
        if self.repeater_end is not None:
            truncated = [m for m in truncated if m.match_index < self.repeater_end]
        return truncated

    def _validate_repeater(self, matches):
        max_match_index = -1
        if matches:
            max_match_index = max(m.match_index for m in matches)
        if max_match_index + 1 < self.repeater_start:
            raise _InvalidChainException

    def chain(self):
        """
        Add patterns chain, using configuration from this chain

        :return:
        :rtype:
        """
        return self._chain.chain()

    def hidden(self, hidden=True):
        """
        Hide chain part results from global chain result

        :param hidden:
        :type hidden:
        :return:
        :rtype:
        """
        self._hidden = hidden
        return self

    @property
    def is_hidden(self):
        """
        Check if the chain part is hidden
        :return:
        :rtype:
        """
        return self._hidden

    def regex(self, *pattern, **kwargs):
        """
        Add re pattern

        :param pattern:
        :type pattern:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return self._chain.regex(*pattern, **kwargs)

    def functional(self, *pattern, **kwargs):
        """
        Add functional pattern

        :param pattern:
        :type pattern:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return self._chain.functional(*pattern, **kwargs)

    def string(self, *pattern, **kwargs):
        """
        Add string pattern

        :param pattern:
        :type pattern:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return self._chain.string(*pattern, **kwargs)

    def close(self):
        """
        Close the chain builder to continue registering other patterns

        :return:
        :rtype:
        """
        return self._chain.close()

    def repeater(self, value):
        """
        Define the repeater of the current chain part.

        :param value:
        :type value:
        :return:
        :rtype:
        """
        try:
            value = int(value)
            self.repeater_start = value
            self.repeater_end = value
            return self
        except ValueError:
            pass
        if value == '+':
            self.repeater_start = 1
            self.repeater_end = None
        if value == '*':
            self.repeater_start = 0
            self.repeater_end = None
        elif value == '?':
            self.repeater_start = 0
            self.repeater_end = 1
        else:
            match = re.match(r'\{\s*(\d*)\s*,?\s*(\d*)\s*\}', value)
            if match:
                start = match.group(1)
                end = match.group(2)
                if start or end:
                    self.repeater_start = int(start) if start else 0
                    self.repeater_end = int(end) if end else None
        return self

    def __repr__(self):
        return f"{self.pattern}({{{self.repeater_start},{self.repeater_end}}})"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Formatter functions to use in patterns.

All those function have last argument as match.value (str).
"""


def formatters(*chained_formatters):
    """
    Chain formatter functions.
    :param chained_formatters:
    :type chained_formatters:
    :return:
    :rtype:
    """

    def formatters_chain(input_string):  # pylint:disable=missing-docstring
        for chained_formatter in chained_formatters:
            input_string = chained_formatter(input_string)
        return input_string

    return formatters_chain


def default_formatter(input_string):
    """
    Default formatter
    :param input_string:
    :return:
    """
    return input_string



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, protected-access, invalid-name, len-as-condition

from .default_rules_module import RuleRemove0
from .. import debug
from ..match import Match
from ..pattern import StringPattern
from ..rebulk import Rebulk


class TestDebug:
    # request.addfinalizer(disable_debug)

    debug.DEBUG = True
    pattern = StringPattern(1, 3, value="es")

    match = Match(1, 3, value="es")
    rule = RuleRemove0()

    input_string = "This is a debug test"
    rebulk = Rebulk().string("debug") \
        .string("is")

    matches = rebulk.matches(input_string)
    debug.DEBUG = False

    @classmethod
    def setup_class(cls):
        debug.DEBUG = True

    @classmethod
    def teardown_class(cls):
        debug.DEBUG = False

    def test_pattern(self):
        assert self.pattern.defined_at.lineno > 0
        assert self.pattern.defined_at.name == 'rebulk.test.test_debug'
        assert self.pattern.defined_at.filename.endswith('test_debug.py')

        assert str(self.pattern.defined_at).startswith('test_debug.py#L')
        assert repr(self.pattern).startswith('<StringPattern@test_debug.py#L')

    def test_match(self):
        assert self.match.defined_at.lineno > 0
        assert self.match.defined_at.name == 'rebulk.test.test_debug'
        assert self.match.defined_at.filename.endswith('test_debug.py')

        assert str(self.match.defined_at).startswith('test_debug.py#L')

    def test_rule(self):
        assert self.rule.defined_at.lineno > 0
        assert self.rule.defined_at.name == 'rebulk.test.test_debug'
        assert self.rule.defined_at.filename.endswith('test_debug.py')

        assert str(self.rule.defined_at).startswith('test_debug.py#L')
        assert repr(self.rule).startswith('<RuleRemove0@test_debug.py#L')

    def test_rebulk(self):
        assert self.rebulk._patterns[0].defined_at.lineno > 0
        assert self.rebulk._patterns[0].defined_at.name == 'rebulk.test.test_debug'
        assert self.rebulk._patterns[0].defined_at.filename.endswith('test_debug.py')

        assert str(self.rebulk._patterns[0].defined_at).startswith('test_debug.py#L')

        assert self.rebulk._patterns[1].defined_at.lineno > 0
        assert self.rebulk._patterns[1].defined_at.name == 'rebulk.test.test_debug'
        assert self.rebulk._patterns[1].defined_at.filename.endswith('test_debug.py')

        assert str(self.rebulk._patterns[1].defined_at).startswith('test_debug.py#L')

        assert self.matches[0].defined_at == self.rebulk._patterns[0].defined_at  # pylint: disable=no-member
        assert self.matches[1].defined_at == self.rebulk._patterns[1].defined_at  # pylint: disable=no-member

    def test_repr(self):
        str(self.matches)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, unbalanced-tuple-unpacking, len-as-condition, no-member

import re
import pytest

from ..pattern import StringPattern, RePattern, FunctionalPattern, REGEX_ENABLED
from ..match import Match

class TestStringPattern:
    """
    Tests for StringPattern matching
    """

    input_string = "An Abyssinian fly playing a Celtic violin was annoyed by trashy flags on " \
                   "which were the Hebrew letter qoph."

    def test_single(self):
        pattern = StringPattern("Celtic")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (28, 34)
        assert matches[0].value == "Celtic"

    def test_repr(self):
        pattern = StringPattern("Celtic")

        assert repr(pattern) == '<StringPattern:(\'Celtic\',)>'

    def test_ignore_case(self):
        pattern = StringPattern("celtic", ignore_case=False)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        pattern = StringPattern("celtic", ignore_case=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert matches[0].value == "Celtic"

    def test_private_names(self):
        pattern = StringPattern("celtic", name="test", private_names=["test"], ignore_case=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert matches[0].private

    def test_ignore_names(self):
        pattern = StringPattern("celtic", name="test", ignore_names=["test"], ignore_case=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

    def test_no_match(self):
        pattern = StringPattern("Python")

        matches = list(pattern.matches(self.input_string))
        assert not matches

    def test_multiple_patterns(self):
        pattern = StringPattern("playing", "annoyed", "Hebrew")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3

        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (18, 25)
        assert matches[0].value == "playing"

        assert isinstance(matches[1], Match)
        assert matches[1].pattern == pattern
        assert matches[1].span == (46, 53)
        assert matches[1].value == "annoyed"

        assert isinstance(matches[2], Match)
        assert matches[2].pattern == pattern
        assert matches[2].span == (88, 94)
        assert matches[2].value == "Hebrew"

    def test_start_end_kwargs(self):
        pattern = StringPattern("Abyssinian", start=20, end=40)
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 0

    def test_matches_kwargs(self):
        pattern = StringPattern("Abyssinian", name="test", value="AB")
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 1
        assert matches[0].name == "test"
        assert matches[0].value == "AB"

    def test_post_processor(self):
        def post_processor(matches, pattern):
            assert len(matches) == 1
            assert isinstance(pattern, StringPattern)

            return []

        pattern = StringPattern("Abyssinian", name="test", value="AB", post_processor=post_processor)
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 0


class TestRePattern:
    """
    Tests for RePattern matching
    """

    input_string = "An Abyssinian fly playing a Celtic violin was annoyed by trashy flags on " \
                   "which were the Hebrew letter qoph."

    def test_single_compiled(self):
        pattern = RePattern(re.compile("Celt.?c"))

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (28, 34)
        assert matches[0].value == "Celtic"

    def test_single_string(self):
        pattern = RePattern("Celt.?c")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (28, 34)
        assert matches[0].value == "Celtic"

    def test_single_kwargs(self):
        pattern = RePattern({"pattern": "celt.?c", "flags": re.IGNORECASE})

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (28, 34)
        assert matches[0].value == "Celtic"

    def test_single_vargs(self):
        pattern = RePattern(("celt.?c", re.IGNORECASE))

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (28, 34)
        assert matches[0].value == "Celtic"

    def test_no_match(self):
        pattern = RePattern("abc.?def")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

    def test_shortcuts(self):
        pattern = RePattern("Celtic-violin", abbreviations=[("-", r"[\W_]+")])

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        pattern = RePattern({"pattern": "celtic-violin", "flags": re.IGNORECASE}, abbreviations=[("-", r"[\W_]+")])

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

    def test_multiple_patterns(self):
        pattern = RePattern("pla.?ing", "ann.?yed", "Heb.?ew")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3

        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (18, 25)
        assert matches[0].value == "playing"

        assert isinstance(matches[1], Match)
        assert matches[1].pattern == pattern
        assert matches[1].span == (46, 53)
        assert matches[1].value == "annoyed"

        assert isinstance(matches[2], Match)
        assert matches[2].pattern == pattern
        assert matches[2].span == (88, 94)
        assert matches[2].value == "Hebrew"

    def test_unnamed_groups(self):
        pattern = RePattern(r"(Celt.?c)\s+(\w+)")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        parent = matches[0]

        assert isinstance(parent, Match)
        assert parent.pattern == pattern
        assert parent.span == (28, 41)
        assert parent.name is None
        assert parent.value == "Celtic violin"

        assert len(parent.children) == 2

        group1, group2 = parent.children

        assert isinstance(group1, Match)
        assert group1.pattern == pattern
        assert group1.span == (28, 34)
        assert group1.name is None
        assert group1.value == "Celtic"
        assert group1.parent == parent

        assert isinstance(group2, Match)
        assert group2.pattern == pattern
        assert group2.span == (35, 41)
        assert group2.name is None
        assert group2.value == "violin"
        assert group2.parent == parent

    def test_named_groups(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        parent = matches[0]

        assert isinstance(parent, Match)
        assert parent.pattern == pattern
        assert parent.span == (28, 41)
        assert parent.name is None
        assert parent.value == "Celtic violin"

        assert len(parent.children) == 2
        group1, group2 = parent.children

        assert isinstance(group1, Match)
        assert group1.pattern == pattern
        assert group1.span == (28, 34)
        assert group1.name == "param1"
        assert group1.value == "Celtic"
        assert group1.parent == parent

        assert isinstance(group2, Match)
        assert group2.pattern == pattern
        assert group2.span == (35, 41)
        assert group2.name == "param2"
        assert group2.value == "violin"
        assert group2.parent == parent

    def test_children(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)", children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 2
        group1, group2 = matches

        assert isinstance(group1, Match)
        assert group1.pattern == pattern
        assert group1.span == (28, 34)
        assert group1.name == "param1"
        assert group1.value == "Celtic"

        assert isinstance(group2, Match)
        assert group2.pattern == pattern
        assert group2.span == (35, 41)
        assert group2.name == "param2"
        assert group2.value == "violin"

    def test_children_parent_private(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)", children=True, private_parent=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3
        parent, group1, group2 = matches

        assert isinstance(group1, Match)
        assert parent.private
        assert parent.pattern == pattern
        assert parent.span == (28, 41)
        assert parent.name is None
        assert parent.value == "Celtic violin"

        assert isinstance(group1, Match)
        assert not group1.private
        assert group1.pattern == pattern
        assert group1.span == (28, 34)
        assert group1.name == "param1"
        assert group1.value == "Celtic"

        assert isinstance(group2, Match)
        assert not group2.private
        assert group2.pattern == pattern
        assert group2.span == (35, 41)
        assert group2.name == "param2"
        assert group2.value == "violin"

    def test_parent_children_private(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)", private_children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3
        parent, group1, group2 = matches

        assert isinstance(group1, Match)
        assert not parent.private
        assert parent.pattern == pattern
        assert parent.span == (28, 41)
        assert parent.name is None
        assert parent.value == "Celtic violin"

        assert isinstance(group1, Match)
        assert group1.private
        assert group1.pattern == pattern
        assert group1.span == (28, 34)
        assert group1.name == "param1"
        assert group1.value == "Celtic"

        assert isinstance(group2, Match)
        assert group2.private
        assert group2.pattern == pattern
        assert group2.span == (35, 41)
        assert group2.name == "param2"
        assert group2.value == "violin"

    def test_every(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)", every=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3
        parent, group1, group2 = matches

        assert isinstance(group1, Match)
        assert not parent.private
        assert parent.pattern == pattern
        assert parent.span == (28, 41)
        assert parent.name is None
        assert parent.value == "Celtic violin"

        assert isinstance(group1, Match)
        assert not group1.private
        assert group1.pattern == pattern
        assert group1.span == (28, 34)
        assert group1.name == "param1"
        assert group1.value == "Celtic"

        assert isinstance(group2, Match)
        assert not group2.private
        assert group2.pattern == pattern
        assert group2.span == (35, 41)
        assert group2.name == "param2"
        assert group2.value == "violin"

    def test_private_names(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)", private_names=["param2"], children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 2
        assert matches[0].name == "param1"
        assert not matches[0].private
        assert matches[1].name == "param2"
        assert matches[1].private

    def test_ignore_names(self):
        pattern = RePattern(r"(?P<param1>Celt.?c)\s+(?P<param2>\w+)", ignore_names=["param2"], children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert matches[0].name == "param1"

    def test_matches_kwargs(self):
        pattern = RePattern("He.rew", name="test", value="HE")
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 1
        assert matches[0].name == "test"
        assert matches[0].value == "HE"

        pattern = RePattern("H(e.)(rew)", name="test", value="HE")
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 1
        assert matches[0].name == "test"
        assert matches[0].value == "HE"

        children = matches[0].children
        assert len(children) == 2
        assert children[0].name == "test"
        assert children[0].value == "HE"

        assert children[1].name == "test"
        assert children[1].value == "HE"

        pattern = RePattern("H(?P<first>e.)(?P<second>rew)", name="test", value="HE")
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 1
        assert matches[0].name == "test"
        assert matches[0].value == "HE"

        children = matches[0].children
        assert len(children) == 2
        assert children[0].name == "first"
        assert children[0].value == "HE"

        assert children[1].name == "second"
        assert children[1].value == "HE"


class TestFunctionalPattern:
    """
    Tests for FunctionalPattern matching
    """

    input_string = "An Abyssinian fly playing a Celtic violin was annoyed by trashy flags on " \
                   "which were the Hebrew letter qoph."

    def test_single_vargs(self):
        def func(input_string):
            i = input_string.find("fly")
            if i > -1:
                return i, i + len("fly"), "fly", "functional"

        pattern = FunctionalPattern(func)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (14, 17)
        assert matches[0].name == "functional"
        assert matches[0].value == "fly"

    def test_single_kwargs(self):
        def func(input_string):
            i = input_string.find("fly")
            if i > -1:
                return {"start": i, "end": i + len("fly"), "name": "functional"}

        pattern = FunctionalPattern(func)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (14, 17)
        assert matches[0].name == "functional"
        assert matches[0].value == "fly"

    def test_multiple_objects(self):
        def func(input_string):
            i = input_string.find("fly")
            matches = []
            if i > -1:
                matches.append((i, i + len("fly"), {'name': "functional"}))
                i = input_string.find("annoyed")
            if i > -1:
                matches.append((i, i + len("annoyed")))
            i = input_string.find("Hebrew")
            if i > -1:
                matches.append({"start": i, "end": i + len("Hebrew")})
            return matches

        pattern = FunctionalPattern(func)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (14, 17)
        assert matches[0].name == "functional"
        assert matches[0].value == "fly"

        assert isinstance(matches[1], Match)
        assert matches[1].pattern == pattern
        assert matches[1].span == (46, 53)
        assert matches[1].value == "annoyed"

        assert isinstance(matches[2], Match)
        assert matches[2].pattern == pattern
        assert matches[2].span == (88, 94)
        assert matches[2].value == "Hebrew"

    def test_multiple_generator(self):
        def func(input_string):
            i = input_string.find("fly")
            if i > -1:
                yield (i, i + len("fly"), {'name': "functional"})
            i = input_string.find("annoyed")
            if i > -1:
                yield (i, i + len("annoyed"))
            i = input_string.find("Hebrew")
            if i > -1:
                yield (i, {"end": i + len("Hebrew")})

        pattern = FunctionalPattern(func)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (14, 17)
        assert matches[0].name == "functional"
        assert matches[0].value == "fly"

        assert isinstance(matches[1], Match)
        assert matches[1].pattern == pattern
        assert matches[1].span == (46, 53)
        assert matches[1].value == "annoyed"

        assert isinstance(matches[2], Match)
        assert matches[2].pattern == pattern
        assert matches[2].span == (88, 94)
        assert matches[2].value == "Hebrew"

    def test_no_match(self):
        pattern = FunctionalPattern(lambda x: None)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

    def test_multiple_patterns(self):
        def playing(input_string):
            i = input_string.find("playing")
            if i > -1:
                return i, i + len("playing")

        def annoyed(input_string):
            i = input_string.find("annoyed")
            if i > -1:
                return i, i + len("annoyed")

        def hebrew(input_string):
            i = input_string.find("Hebrew")
            if i > -1:
                return i, i + len("Hebrew")

        pattern = FunctionalPattern(playing, annoyed, hebrew)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 3

        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (18, 25)
        assert matches[0].value == "playing"

        assert isinstance(matches[1], Match)
        assert matches[1].pattern == pattern
        assert matches[1].span == (46, 53)
        assert matches[1].value == "annoyed"

        assert isinstance(matches[2], Match)
        assert matches[2].pattern == pattern
        assert matches[2].span == (88, 94)
        assert matches[2].value == "Hebrew"

    def test_matches_kwargs(self):
        def playing(input_string):
            i = input_string.find("playing")
            if i > -1:
                return i, i + len("playing")

        pattern = FunctionalPattern(playing, name="test", value="PLAY")
        matches = list(pattern.matches(self.input_string))

        assert len(matches) == 1
        assert matches[0].name == "test"
        assert matches[0].value == "PLAY"


class TestValue:
    """
    Tests for value option
    """

    input_string = "This string contains 1849 a number"

    def test_str_value(self):
        pattern = StringPattern("1849", name="dummy", value="test")

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (21, 25)
        assert matches[0].value == "test"

    def test_dict_child_value(self):
        pattern = RePattern(r"(?P<strParam>cont.?ins)\s+(?P<intParam>\d+)",
                            formatter={'intParam': lambda x: int(x) * 2,
                                       'strParam': lambda x: "really " + x},
                            format_all=True,
                            value={'intParam': 'INT_PARAM_VALUE'})

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        parent = matches[0]
        assert len(parent.children) == 2

        group1, group2 = parent.children

        assert isinstance(group1, Match)
        assert group1.pattern == pattern
        assert group1.span == (12, 20)
        assert group1.value == "really contains"

        assert isinstance(group2, Match)
        assert group2.pattern == pattern
        assert group2.span == (21, 25)
        assert group2.value == 'INT_PARAM_VALUE'

    def test_dict_default_value(self):
        pattern = RePattern(r"(?P<strParam>cont.?ins)\s+(?P<intParam>\d+)",
                            formatter={'intParam': lambda x: int(x) * 2,
                                       'strParam': lambda x: "really " + x},
                            format_all=True,
                            value={'__children__': 'CHILD', 'strParam': 'STR_VALUE', '__parent__': 'PARENT'})

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        parent = matches[0]
        assert parent.value == "PARENT"
        assert len(parent.children) == 2

        group1, group2 = parent.children

        assert isinstance(group1, Match)
        assert group1.pattern == pattern
        assert group1.span == (12, 20)
        assert group1.value == "STR_VALUE"

        assert isinstance(group2, Match)
        assert group2.pattern == pattern
        assert group2.span == (21, 25)
        assert group2.value == "CHILD"


class TestFormatter:
    """
    Tests for formatter option
    """

    input_string = "This string contains 1849 a number"

    def test_single_string(self):
        pattern = StringPattern("1849", name="dummy", formatter=lambda x: int(x) / 2)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (21, 25)
        assert matches[0].value == 1849 / 2

    def test_single_re_no_group(self):
        pattern = RePattern(r"\d+", formatter=lambda x: int(x) * 2)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (21, 25)
        assert matches[0].value == 1849 * 2

    def test_single_re_named_groups(self):
        pattern = RePattern(r"(?P<strParam>cont.?ins)\s+(?P<intParam>\d+)",
                            formatter={'intParam': lambda x: int(x) * 2,
                                       'strParam': lambda x: "really " + x}, format_all=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        parent = matches[0]
        assert len(parent.children) == 2

        group1, group2 = parent.children

        assert isinstance(group1, Match)
        assert group1.pattern == pattern
        assert group1.span == (12, 20)
        assert group1.value == "really contains"

        assert isinstance(group2, Match)
        assert group2.pattern == pattern
        assert group2.span == (21, 25)
        assert group2.value == 1849 * 2

    def test_repeated_captures_option(self):
        pattern = RePattern(r"\[(\d+)\](?:-(\d+))*")

        matches = list(pattern.matches("[02]-03-04-05-06"))
        assert len(matches) == 1

        match = matches[0]
        if REGEX_ENABLED:
            assert len(match.children) == 5
            assert [child.value for child in match.children] == ["02", "03", "04", "05", "06"]
        else:
            assert len(match.children) == 2
            assert [child.value for child in match.children] == ["02", "06"]

            with pytest.raises(NotImplementedError):
                RePattern(r"\[(\d+)\](?:-(\d+))*", repeated_captures=True)

        pattern = RePattern(r"\[(\d+)\](?:-(\d+))*", repeated_captures=False)

        matches = list(pattern.matches("[02]-03-04-05-06"))
        assert len(matches) == 1

        match = matches[0]
        assert len(match.children) == 2
        assert [child.value for child in match.children] == ["02", "06"]

    def test_single_functional(self):
        def digit(input_string):
            i = input_string.find("1849")
            if i > -1:
                return i, i + len("1849")

        pattern = FunctionalPattern(digit, formatter=lambda x: int(x) * 3)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        assert isinstance(matches[0], Match)
        assert matches[0].pattern == pattern
        assert matches[0].span == (21, 25)
        assert matches[0].value == 1849 * 3


class TestValidator:
    """
    Tests for validator option
    """

    input_string = "This string contains 1849 a number"

    @staticmethod
    def true_validator(match):
        return int(match.value) < 1850

    @staticmethod
    def false_validator(match):
        return int(match.value) >= 1850

    def test_single_string(self):
        pattern = StringPattern("1849", name="dummy", validator=self.false_validator)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        pattern = StringPattern("1849", validator=self.true_validator)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

    def test_single_re_no_group(self):
        pattern = RePattern(r"\d+", validator=self.false_validator)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        pattern = RePattern(r"\d+", validator=self.true_validator)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

    def test_single_re_named_groups(self):
        pattern = RePattern(r"(?P<strParam>cont.?ins)\s+(?P<intParam>\d+)",
                            validator={'intParam': self.false_validator}, validate_all=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        pattern = RePattern(r"(?P<strParam>cont.?ins)\s+(?P<intParam>\d+)",
                            validator={'intParam': self.true_validator}, validate_all=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

    def test_validate_all(self):
        pattern = RePattern(r"contains (?P<intParam>\d+)", formatter=int, validator=lambda match: match.value < 100,
                            children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        pattern = RePattern(r"contains (?P<intParam>\d+)", formatter=int, validator=lambda match: match.value > 100,
                            children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

        def invalid_func(match):
            if match.name == 'intParam':
                return True
            return match.value.startswith('abc')

        pattern = RePattern(r"contains (?P<intParam>\d+)", formatter=int, validator=invalid_func, validate_all=True,
                            children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        def func(match):
            if match.name == 'intParam':
                return True
            return match.value.startswith('contains')

        pattern = RePattern(r"contains (?P<intParam>\d+)", formatter=int, validator=func, validate_all=True,
                            children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1

    def test_format_all(self):
        pattern = RePattern(r"contains (?P<intParam>\d+)", formatter=int,
                            children=True)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1
        for match in matches:
            assert match.value is not None

        with pytest.raises(ValueError):
            pattern = RePattern(r"contains (?P<intParam>\d+)", formatter=int, format_all=True)
            matches = list(pattern.matches(self.input_string))
            for match in matches:
                assert match.value is not None

    def test_single_functional(self):
        def digit(input_string):
            i = input_string.find("1849")
            if i > -1:
                return i, i + len("1849")

        pattern = FunctionalPattern(digit, validator=self.false_validator)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 0

        pattern = FunctionalPattern(digit, validator=self.true_validator)

        matches = list(pattern.matches(self.input_string))
        assert len(matches) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Introspect rebulk object to retrieve capabilities.
"""
from abc import ABCMeta, abstractmethod
from collections import defaultdict

from .pattern import StringPattern, RePattern, FunctionalPattern
from .utils import extend_safe


class Description(metaclass=ABCMeta):
    """
    Abstract class for a description.
    """
    @property
    @abstractmethod
    def properties(self):  # pragma: no cover
        """
        Properties of described object.
        :return: all properties that described object can generate grouped by name.
        :rtype: dict
        """


class PatternDescription(Description):
    """
    Description of a pattern.
    """
    def __init__(self, pattern):  # pylint:disable=too-many-branches
        self.pattern = pattern
        self._properties = defaultdict(list)

        if pattern.properties:
            for key, values in pattern.properties.items():
                extend_safe(self._properties[key], values)
        elif 'value' in pattern.match_options:
            self._properties[pattern.name].append(pattern.match_options['value'])
        elif isinstance(pattern, StringPattern):
            extend_safe(self._properties[pattern.name], pattern.patterns)
        elif isinstance(pattern, RePattern):
            if pattern.name and pattern.name not in pattern.private_names:
                extend_safe(self._properties[pattern.name], [None])
            if not pattern.private_children:
                for regex_pattern in pattern.patterns:
                    for group_name, values in regex_pattern.groupindex.items():
                        if group_name not in pattern.private_names:
                            extend_safe(self._properties[group_name], [None])
        elif isinstance(pattern, FunctionalPattern):
            if pattern.name and pattern.name not in pattern.private_names:
                extend_safe(self._properties[pattern.name], [None])


    @property
    def properties(self):
        """
        Properties for this rule.
        :return:
        :rtype: dict
        """
        return self._properties


class RuleDescription(Description):
    """
    Description of a rule.
    """
    def __init__(self, rule):
        self.rule = rule

        self._properties = defaultdict(list)

        if rule.properties:
            for key, values in rule.properties.items():
                extend_safe(self._properties[key], values)

    @property
    def properties(self):
        """
        Properties for this rule.
        :return:
        :rtype: dict
        """
        return self._properties


class Introspection(Description):
    """
    Introspection results.
    """
    def __init__(self, rebulk, context=None):
        self.patterns = [PatternDescription(pattern) for pattern in rebulk.effective_patterns(context)
                         if not pattern.private and not pattern.marker]
        self.rules = [RuleDescription(rule) for rule in rebulk.effective_rules(context)]

    @property
    def properties(self):
        """
        Properties for Introspection results.
        :return:
        :rtype:
        """
        properties = defaultdict(list)
        for pattern in self.patterns:
            for key, values in pattern.properties.items():
                extend_safe(properties[key], values)
        for rule in self.rules:
            for key, values in rule.properties.items():
                extend_safe(properties[key], values)
        return properties


def introspect(rebulk, context=None):
    """
    Introspect a Rebulk instance to grab defined objects and properties that can be generated.
    :param rebulk:
    :type rebulk: Rebulk
    :param context:
    :type context:
    :return: Introspection instance
    :rtype: Introspection
    """
    return Introspection(rebulk, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Abstract pattern class definition along with various implementations (regexp, string, functional)
"""
# pylint: disable=super-init-not-called,wrong-import-position

from abc import ABCMeta, abstractmethod

from . import debug
from .formatters import default_formatter
from .loose import call, ensure_list, ensure_dict
from .match import Match
from .remodule import re, REGEX_ENABLED
from .utils import find_all, is_iterable, get_first_defined
from .validators import allways_true


class BasePattern(metaclass=ABCMeta):
    """
    Base class for Pattern like objects
    """

    @abstractmethod
    def matches(self, input_string, context=None, with_raw_matches=False):
        """
        Computes all matches for a given input

        :param input_string: the string to parse
        :type input_string: str
        :param context: the context
        :type context: dict
        :param with_raw_matches: should return details
        :type with_raw_matches: dict
        :return: matches based on input_string for this pattern
        :rtype: iterator[Match]
        """


class Pattern(BasePattern, metaclass=ABCMeta):
    """
    Definition of a particular pattern to search for.
    """

    def __init__(self, name=None, tags=None, formatter=None, value=None, validator=None, children=False, every=False,
                 private_parent=False, private_children=False, private=False, private_names=None, ignore_names=None,
                 marker=False, format_all=False, validate_all=False, disabled=lambda context: False, log_level=None,
                 properties=None, post_processor=None, pre_match_processor=None, post_match_processor=None, **kwargs):
        """
        :param name: Name of this pattern
        :type name: str
        :param tags: List of tags related to this pattern
        :type tags: list[str]
        :param formatter: dict (name, func) of formatter to use with this pattern. name is the match name to support,
        and func a function(input_string) that returns the formatted string. A single formatter function can also be
        passed as a shortcut for {None: formatter}. The returned formatted string with be set in Match.value property.
        :type formatter: dict[str, func] || func
        :param value: dict (name, value) of value to use with this pattern. name is the match name to support,
        and value an object for the match value. A single object value can also be
        passed as a shortcut for {None: value}. The value with be set in Match.value property.
        :type value: dict[str, object] || object
        :param validator: dict (name, func) of validator to use with this pattern. name is the match name to support,
        and func a function(match) that returns the a boolean. A single validator function can also be
        passed as a shortcut for {None: validator}. If return value is False, match will be ignored.
        :param children: generates children instead of parent
        :type children: bool
        :param every: generates both parent and children.
        :type every: bool
        :param private: flag this pattern as beeing private.
        :type private: bool
        :param private_parent: force return of parent and flag parent matches as private.
        :type private_parent: bool
        :param private_children: force return of children and flag children matches as private.
        :type private_children: bool
        :param private_names: force return of named matches as private.
        :type private_names: bool
        :param ignore_names: drop some named matches after validation.
        :type ignore_names: bool
        :param marker: flag this pattern as beeing a marker.
        :type private: bool
        :param format_all if True, pattern will format every match in the hierarchy (even match not yield).
        :type format_all: bool
        :param validate_all if True, pattern will validate every match in the hierarchy (even match not yield).
        :type validate_all: bool
        :param disabled: if True, this pattern is disabled. Can also be a function(context).
        :type disabled: bool|function
        :param log_lvl: Log level associated to this pattern
        :type log_lvl: int
        :param post_processor: Post processing function
        :type post_processor: func
        :param pre_match_processor: Pre match processing function
        :type pre_match_processor: func
        :param post_match_processor: Post match processing function
        :type post_match_processor: func
        """
        # pylint:disable=too-many-locals,unused-argument
        self.name = name
        self.tags = ensure_list(tags)
        self.formatters, self._default_formatter = ensure_dict(formatter, default_formatter)
        self.values, self._default_value = ensure_dict(value, None)
        self.validators, self._default_validator = ensure_dict(validator, allways_true)
        self.every = every
        self.children = children
        self.private = private
        self.private_names = private_names if private_names else []
        self.ignore_names = ignore_names if ignore_names else []
        self.private_parent = private_parent
        self.private_children = private_children
        self.marker = marker
        self.format_all = format_all
        self.validate_all = validate_all
        if not callable(disabled):
            self.disabled = lambda context: disabled
        else:
            self.disabled = disabled
        self._log_level = log_level
        self._properties = properties
        self.defined_at = debug.defined_at()
        if not callable(post_processor):
            self.post_processor = None
        else:
            self.post_processor = post_processor
        if not callable(pre_match_processor):
            self.pre_match_processor = None
        else:
            self.pre_match_processor = pre_match_processor
        if not callable(post_match_processor):
            self.post_match_processor = None
        else:
            self.post_match_processor = post_match_processor

    @property
    def log_level(self):
        """
        Log level for this pattern.
        :return:
        :rtype:
        """
        return self._log_level if self._log_level is not None else debug.LOG_LEVEL

    def matches(self, input_string, context=None, with_raw_matches=False):
        """
        Computes all matches for a given input

        :param input_string: the string to parse
        :type input_string: str
        :param context: the context
        :type context: dict
        :param with_raw_matches: should return details
        :type with_raw_matches: dict
        :return: matches based on input_string for this pattern
        :rtype: iterator[Match]
        """
        # pylint: disable=too-many-branches

        matches = []
        raw_matches = []

        for pattern in self.patterns:
            match_index = 0
            for match in self._match(pattern, input_string, context):
                raw_matches.append(match)
                matches.extend(self._process_matches(match, match_index))
                match_index += 1

        matches = self._post_process_matches(matches)

        if with_raw_matches:
            return matches, raw_matches
        return matches

    @property
    def _should_include_children(self):
        """
        Check if children matches from this pattern should be included in matches results.
        :param match:
        :type match:
        :return:
        :rtype:
        """
        return self.children or self.every

    @property
    def _should_include_parent(self):
        """
        Check is a match from this pattern should be included in matches results.
        :param match:
        :type match:
        :return:
        :rtype:
        """
        return not self.children or self.every

    @staticmethod
    def _match_config_property_keys(match, child=False):
        if match.name:
            yield match.name
        if child:
            yield '__children__'
        else:
            yield '__parent__'
        yield None

    @staticmethod
    def _process_match_index(match, match_index):
        """
        Process match index from this pattern process state.

        :param match:
        :return:
        """
        match.match_index = match_index

    def _process_match_private(self, match, child=False):
        """
        Process match privacy from this pattern configuration.

        :param match:
        :param child:
        :return:
        """

        if match.name and match.name in self.private_names or \
                not child and self.private_parent or \
                child and self.private_children:
            match.private = True

    def _process_match_value(self, match, child=False):
        """
        Process match value from this pattern configuration.
        :param match:
        :return:
        """
        keys = self._match_config_property_keys(match, child=child)
        pattern_value = get_first_defined(self.values, keys, self._default_value)
        if pattern_value:
            match.value = pattern_value

    def _process_match_formatter(self, match, child=False):
        """
        Process match formatter from this pattern configuration.

        :param match:
        :return:
        """
        included = self._should_include_children if child else self._should_include_parent
        if included or self.format_all:
            keys = self._match_config_property_keys(match, child=child)
            match.formatter = get_first_defined(self.formatters, keys, self._default_formatter)

    def _process_match_validator(self, match, child=False):
        """
        Process match validation from this pattern configuration.

        :param match:
        :return: True if match is validated by the configured validator, False otherwise.
        """
        included = self._should_include_children if child else self._should_include_parent
        if included or self.validate_all:
            keys = self._match_config_property_keys(match, child=child)
            validator = get_first_defined(self.validators, keys, self._default_validator)
            if validator and not validator(match):
                return False
        return True

    def _process_match(self, match, match_index, child=False):
        """
        Process match from this pattern by setting all properties from defined configuration
        (index, private, value, formatter, validator, ...).

        :param match:
        :type match:
        :return: True if match is validated by the configured validator, False otherwise.
        :rtype:
        """
        self._process_match_index(match, match_index)
        self._process_match_private(match, child)
        self._process_match_value(match, child)
        self._process_match_formatter(match, child)
        return self._process_match_validator(match, child)

    @staticmethod
    def _process_match_processor(match, processor):
        if processor:
            ret = processor(match)
            if ret is not None:
                return ret
        return match

    def _process_matches(self, match, match_index):
        """
        Process and generate all matches for the given unprocessed match.
        :param match:
        :param match_index:
        :return: Process and dispatched matches.
        """
        match = self._process_match_processor(match, self.pre_match_processor)
        if not match:
            return

        if not self._process_match(match, match_index):
            return

        for child in match.children:
            if not self._process_match(child, match_index, child=True):
                return

        match = self._process_match_processor(match, self.post_match_processor)
        if not match:
            return

        if (self._should_include_parent or self.private_parent) and match.name not in self.ignore_names:
            yield match
        if self._should_include_children or self.private_children:
            children = [x for x in match.children if x.name not in self.ignore_names]
            for child in children:
                yield child

    def _post_process_matches(self, matches):
        """
        Post process matches with user defined function
        :param matches:
        :type matches:
        :return:
        :rtype:
        """
        if self.post_processor:
            return self.post_processor(matches, self)
        return matches

    @property
    @abstractmethod
    def patterns(self):  # pragma: no cover
        """
        List of base patterns defined

        :return: A list of base patterns
        :rtype: list
        """

    @property
    def properties(self):
        """
        Properties names and values that can ben retrieved by this pattern.
        :return:
        :rtype:
        """
        if self._properties:
            return self._properties
        return {}

    @property
    @abstractmethod
    def match_options(self):  # pragma: no cover
        """
        dict of default options for generated Match objects

        :return: **options to pass to Match constructor
        :rtype: dict
        """

    @abstractmethod
    def _match(self, pattern, input_string, context=None):  # pragma: no cover
        """
        Computes all unprocess matches for a given pattern and input.

        :param pattern: the pattern to use
        :param input_string: the string to parse
        :type input_string: str
        :param context: the context
        :type context: dict
        :return: matches based on input_string for this pattern
        :rtype: iterator[Match]
        """

    def __repr__(self):
        defined = ""
        if self.defined_at:
            defined = f"@{self.defined_at}"
        return f"<{self.__class__.__name__}{defined}:{self.__repr__patterns__}>"

    @property
    def __repr__patterns__(self):
        return self.patterns


class StringPattern(Pattern):
    """
    Definition of one or many strings to search for.
    """

    def __init__(self, *patterns, **kwargs):
        super().__init__(**kwargs)
        self._patterns = patterns
        self._kwargs = kwargs
        self._match_kwargs = filter_match_kwargs(kwargs)

    @property
    def patterns(self):
        return self._patterns

    @property
    def match_options(self):
        return self._match_kwargs

    def _match(self, pattern, input_string, context=None):
        for index in find_all(input_string, pattern, **self._kwargs):
            match = Match(index, index + len(pattern), pattern=self, input_string=input_string, **self._match_kwargs)
            if match:
                yield match


class RePattern(Pattern):
    """
    Definition of one or many regular expression pattern to search for.
    """

    def __init__(self, *patterns, **kwargs):
        super().__init__(**kwargs)
        self.repeated_captures = REGEX_ENABLED
        if 'repeated_captures' in kwargs:
            self.repeated_captures = kwargs.get('repeated_captures')
        if self.repeated_captures and not REGEX_ENABLED:  # pragma: no cover
            raise NotImplementedError("repeated_capture is available only with regex module.")
        self.abbreviations = kwargs.get('abbreviations', [])
        self._kwargs = kwargs
        self._match_kwargs = filter_match_kwargs(kwargs)
        self._children_match_kwargs = filter_match_kwargs(kwargs, children=True)
        self._patterns = []
        for pattern in patterns:
            if isinstance(pattern, str):
                if self.abbreviations and pattern:
                    for key, replacement in self.abbreviations:
                        pattern = pattern.replace(key, replacement)
                pattern = call(re.compile, pattern, **self._kwargs)
            elif isinstance(pattern, dict):
                if self.abbreviations and 'pattern' in pattern:
                    for key, replacement in self.abbreviations:
                        pattern['pattern'] = pattern['pattern'].replace(key, replacement)
                pattern = re.compile(**pattern)
            elif hasattr(pattern, '__iter__'):
                pattern = re.compile(*pattern)
            self._patterns.append(pattern)

    @property
    def patterns(self):
        return self._patterns

    @property
    def __repr__patterns__(self):
        return [pattern.pattern for pattern in self.patterns]

    @property
    def match_options(self):
        return self._match_kwargs

    def _match(self, pattern, input_string, context=None):
        names = dict((v, k) for k, v in pattern.groupindex.items())
        for match_object in pattern.finditer(input_string):
            start = match_object.start()
            end = match_object.end()
            main_match = Match(start, end, pattern=self, input_string=input_string, **self._match_kwargs)

            if pattern.groups:
                for i in range(1, pattern.groups + 1):
                    name = names.get(i, main_match.name)
                    if self.repeated_captures:
                        for start, end in match_object.spans(i):
                            child_match = Match(start, end, name=name, parent=main_match, pattern=self,
                                                input_string=input_string, **self._children_match_kwargs)
                            if child_match:
                                main_match.children.append(child_match)
                    else:
                        start, end = match_object.span(i)
                        if start > -1 and end > -1:
                            child_match = Match(start, end, name=name, parent=main_match, pattern=self,
                                                input_string=input_string, **self._children_match_kwargs)
                            if child_match:
                                main_match.children.append(child_match)

            if main_match:
                yield main_match


class FunctionalPattern(Pattern):
    """
    Definition of one or many functional pattern to search for.
    """

    def __init__(self, *patterns, **kwargs):
        super().__init__(**kwargs)
        self._patterns = patterns
        self._kwargs = kwargs
        self._match_kwargs = filter_match_kwargs(kwargs)

    @property
    def patterns(self):
        return self._patterns

    @property
    def match_options(self):
        return self._match_kwargs

    def _match(self, pattern, input_string, context=None):
        ret = call(pattern, input_string, context, **self._kwargs)
        if ret:
            if not is_iterable(ret) or isinstance(ret, dict) \
                    or (is_iterable(ret) and hasattr(ret, '__getitem__') and isinstance(ret[0], int)):
                args_iterable = [ret]
            else:
                args_iterable = ret
            for args in args_iterable:
                if isinstance(args, dict):
                    options = args
                    options.pop('input_string', None)
                    options.pop('pattern', None)
                    if self._match_kwargs:
                        options = self._match_kwargs.copy()
                        options.update(args)
                    match = Match(pattern=self, input_string=input_string, **options)
                    if match:
                        yield match
                else:
                    kwargs = self._match_kwargs
                    if isinstance(args[-1], dict):
                        kwargs = dict(kwargs)
                        kwargs.update(args[-1])
                        args = args[:-1]
                    match = Match(*args, pattern=self, input_string=input_string, **kwargs)
                    if match:
                        yield match


def filter_match_kwargs(kwargs, children=False):
    """
    Filters out kwargs for Match construction

    :param kwargs:
    :type kwargs: dict
    :param children:
    :type children: Flag to filter children matches
    :return: A filtered dict
    :rtype: dict
    """
    kwargs = kwargs.copy()
    for key in ('pattern', 'start', 'end', 'parent', 'formatter', 'value'):
        if key in kwargs:
            del kwargs[key]
    if children:
        for key in ('name',):
            if key in kwargs:
                del kwargs[key]
    return kwargs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, invalid-name, len-as-condition
from rebulk.rules import Rule, RemoveMatch, CustomRule


class RemoveAllButLastYear(Rule):
    consequence = RemoveMatch
    def when(self, matches, context):
        entries = matches.named('year')
        return entries[:-1]


class PrefixedSuffixedYear(CustomRule):
    def when(self, matches, context):
        toRemove = []
        years = matches.named('year')
        for year in years:
            if not matches.previous(year, lambda p: p.name == 'yearPrefix') and \
                   not matches.next(year, lambda n: n.name == 'yearSuffix'):
                toRemove.append(year)
        return toRemove

    def then(self, matches, when_response, context):
        for to_remove in when_response:
            matches.remove(to_remove)


class PrefixedSuffixedYearNoLambda(Rule):
    consequence = RemoveMatch
    def when(self, matches, context):
        toRemove = []
        years = matches.named('year')
        for year in years:
            if not [m for m in matches.previous(year) if m.name == 'yearPrefix'] and \
                    not [m for m in matches.next(year) if m.name == 'yearSuffix']:
                toRemove.append(year)
        return toRemove



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, no-member, len-as-condition, cyclic-import
import re
from functools import partial

from rebulk.pattern import FunctionalPattern, StringPattern, RePattern
from ..rebulk import Rebulk
from ..validators import chars_surround


def test_chain_close():
    rebulk = Rebulk()
    ret = rebulk.chain().close()

    assert ret == rebulk
    assert len(rebulk.effective_patterns()) == 1


def test_build_chain():
    rebulk = Rebulk()

    def digit(input_string):
        i = input_string.find("1849")
        if i > -1:
            return i, i + len("1849")

    ret = rebulk.chain() \
        .functional(digit) \
        .string("test").repeater(2) \
        .string("x").repeater('{1,3}') \
        .string("optional").repeater('?') \
        .regex("f?x").repeater('+') \
        .close()

    assert ret == rebulk
    assert len(rebulk.effective_patterns()) == 1

    chain = rebulk.effective_patterns()[0]

    assert len(chain.parts) == 5

    assert isinstance(chain.parts[0].pattern, FunctionalPattern)
    assert chain.parts[0].repeater_start == 1
    assert chain.parts[0].repeater_end == 1

    assert isinstance(chain.parts[1].pattern, StringPattern)
    assert chain.parts[1].repeater_start == 2
    assert chain.parts[1].repeater_end == 2

    assert isinstance(chain.parts[2].pattern, StringPattern)
    assert chain.parts[2].repeater_start == 1
    assert chain.parts[2].repeater_end == 3

    assert isinstance(chain.parts[3].pattern, StringPattern)
    assert chain.parts[3].repeater_start == 0
    assert chain.parts[3].repeater_end == 1

    assert isinstance(chain.parts[4].pattern, RePattern)
    assert chain.parts[4].repeater_start == 1
    assert chain.parts[4].repeater_end is None


def test_chain_defaults():
    rebulk = Rebulk()
    rebulk.defaults(validator=lambda x: x.value.startswith('t'), ignore_names=['testIgnore'], children=True)

    rebulk.chain() \
        .regex("(?P<test>test)") \
        .regex(" ").repeater("*") \
        .regex("(?P<best>best)") \
        .regex(" ").repeater("*") \
        .regex("(?P<testIgnore>testIgnore)")
    matches = rebulk.matches("test best testIgnore")

    assert len(matches) == 1
    assert matches[0].name == "test"


def test_chain_with_validators():
    def chain_validator(match):
        return match.value.startswith('t') and match.value.endswith('t')

    def default_validator(match):
        return match.value.startswith('t') and match.value.endswith('g')

    def custom_validator(match):
        return match.value.startswith('b') and match.value.endswith('t')

    rebulk = Rebulk()
    rebulk.defaults(children=True, validator=default_validator)

    rebulk.chain(validate_all=True, validator={'__parent__': chain_validator}) \
        .regex("(?P<test>testing)", validator=default_validator).repeater("+") \
        .regex(" ").repeater("+") \
        .regex("(?P<best>best)", validator=custom_validator).repeater("+")
    matches = rebulk.matches("some testing best end")

    assert len(matches) == 2
    assert matches[0].name == "test"
    assert matches[1].name == "best"


def test_matches_docs():
    rebulk = Rebulk().regex_defaults(flags=re.IGNORECASE) \
        .defaults(children=True, formatter={'episode': int, 'version': int}) \
        .chain() \
        .regex(r'e(?P<episode>\d{1,4})').repeater(1) \
        .regex(r'v(?P<version>\d+)').repeater('?') \
        .regex(r'[ex-](?P<episode>\d{1,4})').repeater('*') \
        .close()  # .repeater(1) could be omitted as it's the default behavior

    result = rebulk.matches("This is E14v2-15-16-17").to_dict()  # converts matches to dict

    assert 'episode' in result
    assert result['episode'] == [14, 15, 16, 17]
    assert 'version' in result
    assert result['version'] == 2


def test_matches():
    rebulk = Rebulk()

    def digit(input_string):
        i = input_string.find("1849")
        if i > -1:
            return i, i + len("1849")

    input_string = "1849testtestxxfixfux_foxabc1849testtestxoptionalfoxabc"

    chain = rebulk.chain() \
        .functional(digit) \
        .string("test").hidden().repeater(2) \
        .string("x").hidden().repeater('{1,3}') \
        .string("optional").hidden().repeater('?') \
        .regex("f.?x", name='result').repeater('+') \
        .close()

    matches = chain.matches(input_string)

    assert len(matches) == 2
    children = matches[0].children

    assert children[0].value == '1849'
    assert children[1].value == 'fix'
    assert children[2].value == 'fux'

    children = matches[1].children
    assert children[0].value == '1849'
    assert children[1].value == 'fox'

    input_string = "_1850testtestxoptionalfoxabc"
    matches = chain.matches(input_string)

    assert len(matches) == 0

    input_string = "_1849testtesttesttestxoptionalfoxabc"
    matches = chain.matches(input_string)

    assert len(matches) == 0

    input_string = "_1849testtestxxxxoptionalfoxabc"
    matches = chain.matches(input_string)

    assert len(matches) == 0

    input_string = "_1849testtestoptionalfoxabc"
    matches = chain.matches(input_string)

    assert len(matches) == 0

    input_string = "_1849testtestxoptionalabc"
    matches = chain.matches(input_string)

    assert len(matches) == 0

    input_string = "_1849testtestxoptionalfaxabc"
    matches = chain.matches(input_string)

    assert len(matches) == 1
    children = matches[0].children

    assert children[0].value == '1849'
    assert children[1].value == 'fax'


def test_matches_2():
    rebulk = Rebulk() \
        .regex_defaults(flags=re.IGNORECASE) \
        .defaults(children=True, formatter={'episode': int, 'version': int}) \
        .chain() \
        .regex(r'e(?P<episode>\d{1,4})') \
        .regex(r'v(?P<version>\d+)').repeater('?') \
        .regex(r'[ex-](?P<episode>\d{1,4})').repeater('*') \
        .close()

    matches = rebulk.matches("This is E14v2-15E16x17")
    assert len(matches) == 5

    assert matches[0].name == 'episode'
    assert matches[0].value == 14

    assert matches[1].name == 'version'
    assert matches[1].value == 2

    assert matches[2].name == 'episode'
    assert matches[2].value == 15

    assert matches[3].name == 'episode'
    assert matches[3].value == 16

    assert matches[4].name == 'episode'
    assert matches[4].value == 17


def test_matches_3():
    alt_dash = (r'@', r'[\W_]')  # abbreviation

    match_names = ['season', 'episode']
    other_names = ['screen_size', 'video_codec', 'audio_codec', 'audio_channels', 'container', 'date']

    rebulk = Rebulk()
    rebulk.defaults(formatter={'season': int, 'episode': int},
                    tags=['SxxExx'],
                    abbreviations=[alt_dash],
                    private_names=['episodeSeparator', 'seasonSeparator'],
                    children=True,
                    private_parent=True,
                    conflict_solver=lambda match, other: match
                    if match.name in match_names and other.name in other_names
                    else '__default__')

    rebulk.chain() \
        .defaults(children=True, private_parent=True) \
        .regex(r'(?P<season>\d+)@?x@?(?P<episode>\d+)') \
        .regex(r'(?P<episodeSeparator>x|-|\+|&)(?P<episode>\d+)').repeater('*') \
        .close() \
        .chain() \
        .defaults(children=True, private_parent=True) \
        .regex(r'S(?P<season>\d+)@?(?:xE|Ex|E|x)@?(?P<episode>\d+)') \
        .regex(r'(?:(?P<episodeSeparator>xE|Ex|E|x|-|\+|&)(?P<episode>\d+))').repeater('*') \
        .close() \
        .chain() \
        .defaults(children=True, private_parent=True) \
        .regex(r'S(?P<season>\d+)') \
        .regex(r'(?P<seasonSeparator>S|-|\+|&)(?P<season>\d+)').repeater('*')

    matches = rebulk.matches("test-01x02-03")
    assert len(matches) == 3

    assert matches[0].name == 'season'
    assert matches[0].value == 1

    assert matches[1].name == 'episode'
    assert matches[1].value == 2

    assert matches[2].name == 'episode'
    assert matches[2].value == 3

    matches = rebulk.matches("test-S01E02-03")

    assert len(matches) == 3
    assert matches[0].name == 'season'
    assert matches[0].value == 1

    assert matches[1].name == 'episode'
    assert matches[1].value == 2

    assert matches[2].name == 'episode'
    assert matches[2].value == 3

    matches = rebulk.matches("test-S01-02-03-04")

    assert len(matches) == 4
    assert matches[0].name == 'season'
    assert matches[0].value == 1

    assert matches[1].name == 'season'
    assert matches[1].value == 2

    assert matches[2].name == 'season'
    assert matches[2].value == 3

    assert matches[3].name == 'season'
    assert matches[3].value == 4


def test_matches_4():
    seps_surround = partial(chars_surround, " ")

    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)
    rebulk.defaults(validate_all=True, children=True)
    rebulk.defaults(private_names=['episodeSeparator', 'seasonSeparator'], private_parent=True)

    rebulk.chain(validator={'__parent__': seps_surround}, formatter={'episode': int, 'version': int}) \
        .defaults(formatter={'episode': int, 'version': int}) \
        .regex(r'e(?P<episode>\d{1,4})') \
        .regex(r'v(?P<version>\d+)').repeater('?') \
        .regex(r'(?P<episodeSeparator>e|x|-)(?P<episode>\d{1,4})').repeater('*')

    matches = rebulk.matches("Some Series E01E02E03")
    assert len(matches) == 3

    assert matches[0].value == 1
    assert matches[1].value == 2
    assert matches[2].value == 3


def test_matches_5():
    seps_surround = partial(chars_surround, " ")

    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)

    rebulk.chain(private_names=['episodeSeparator', 'seasonSeparator'], validate_all=True,
                 validator={'__parent__': seps_surround}, children=True, private_parent=True,
                 formatter={'episode': int, 'version': int}) \
        .defaults(children=True, private_parent=True) \
        .regex(r'e(?P<episode>\d{1,4})') \
        .regex(r'v(?P<version>\d+)').repeater('?') \
        .regex(r'(?P<episodeSeparator>e|x|-)(?P<episode>\d{1,4})').repeater('{2,3}')

    matches = rebulk.matches("Some Series E01E02E03")
    assert len(matches) == 3

    matches = rebulk.matches("Some Series E01E02")
    assert len(matches) == 0

    matches = rebulk.matches("Some Series E01E02E03E04E05E06")  # Parent can't be validated, so no results at all
    assert len(matches) == 0


def test_matches_6():
    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)
    rebulk.defaults(private_names=['episodeSeparator', 'seasonSeparator'], validate_all=True,
                    validator=None, children=True, private_parent=True)

    rebulk.chain(formatter={'episode': int, 'version': int}) \
        .defaults(children=True, private_parent=True) \
        .regex(r'e(?P<episode>\d{1,4})') \
        .regex(r'v(?P<version>\d+)').repeater('?') \
        .regex(r'(?P<episodeSeparator>e|x|-)(?P<episode>\d{1,4})').repeater('{2,3}')

    matches = rebulk.matches("Some Series E01E02E03")
    assert len(matches) == 3

    matches = rebulk.matches("Some Series E01E02")
    assert len(matches) == 0

    matches = rebulk.matches("Some Series E01E02E03E04E05E06")  # No validator on parent, so it should give 4 episodes.
    assert len(matches) == 4


def test_matches_7():
    seps_surround = partial(chars_surround, ' .-/')
    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)
    rebulk.defaults(children=True, private_parent=True)

    rebulk.chain(). \
        regex(r'S(?P<season>\d+)', validate_all=True, validator={'__parent__': seps_surround}). \
        regex(r'[ -](?P<season>\d+)', validator=seps_surround).repeater('*')

    matches = rebulk.matches("Some S01")
    assert len(matches) == 1
    matches[0].value = 1

    matches = rebulk.matches("Some S01-02")
    assert len(matches) == 2
    matches[0].value = 1
    matches[1].value = 2

    matches = rebulk.matches("programs4/Some S01-02")
    assert len(matches) == 2
    matches[0].value = 1
    matches[1].value = 2

    matches = rebulk.matches("programs4/SomeS01middle.S02-03.andS04here")
    assert len(matches) == 2
    matches[0].value = 2
    matches[1].value = 3

    matches = rebulk.matches("Some 02.and.S04-05.here")
    assert len(matches) == 2
    matches[0].value = 4
    matches[1].value = 5


def test_chain_breaker():
    def chain_breaker(matches):
        seasons = matches.named('season')
        if len(seasons) > 1:
            if seasons[-1].value - seasons[-2].value > 10:
                return True
        return False

    seps_surround = partial(chars_surround, ' .-/')
    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)
    rebulk.defaults(children=True, private_parent=True, formatter={'season': int})

    rebulk.chain(chain_breaker=chain_breaker). \
        regex(r'S(?P<season>\d+)', validate_all=True, validator={'__parent__': seps_surround}). \
        regex(r'[ -](?P<season>\d+)', validator=seps_surround).repeater('*')

    matches = rebulk.matches("Some S01-02-03-50-51")
    assert len(matches) == 3
    matches[0].value = 1
    matches[1].value = 2
    matches[2].value = 3


def test_chain_breaker_defaults():
    def chain_breaker(matches):
        seasons = matches.named('season')
        if len(seasons) > 1:
            if seasons[-1].value - seasons[-2].value > 10:
                return True
        return False

    seps_surround = partial(chars_surround, ' .-/')
    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)
    rebulk.defaults(chain_breaker=chain_breaker, children=True, private_parent=True, formatter={'season': int})

    rebulk.chain(). \
        regex(r'S(?P<season>\d+)', validate_all=True, validator={'__parent__': seps_surround}). \
        regex(r'[ -](?P<season>\d+)', validator=seps_surround).repeater('*')

    matches = rebulk.matches("Some S01-02-03-50-51")
    assert len(matches) == 3
    matches[0].value = 1
    matches[1].value = 2
    matches[2].value = 3


def test_chain_breaker_defaults2():
    def chain_breaker(matches):
        seasons = matches.named('season')
        if len(seasons) > 1:
            if seasons[-1].value - seasons[-2].value > 10:
                return True
        return False

    seps_surround = partial(chars_surround, ' .-/')
    rebulk = Rebulk()
    rebulk.regex_defaults(flags=re.IGNORECASE)
    rebulk.chain_defaults(chain_breaker=chain_breaker)
    rebulk.defaults(children=True, private_parent=True, formatter={'season': int})

    rebulk.chain(). \
        regex(r'S(?P<season>\d+)', validate_all=True, validator={'__parent__': seps_surround}). \
        regex(r'[ -](?P<season>\d+)', validator=seps_surround).repeater('*')

    matches = rebulk.matches("Some S01-02-03-50-51")
    assert len(matches) == 3
    matches[0].value = 1
    matches[1].value = 2
    matches[2].value = 3



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Various utilities functions
"""
from collections.abc import MutableSet

from types import GeneratorType


def find_all(string, sub, start=None, end=None, ignore_case=False, **kwargs):
    """
    Return all indices in string s where substring sub is
    found, such that sub is contained in the slice s[start:end].

    >>> list(find_all('The quick brown fox jumps over the lazy dog', 'fox'))
    [16]

    >>> list(find_all('The quick brown fox jumps over the lazy dog', 'mountain'))
    []

    >>> list(find_all('The quick brown fox jumps over the lazy dog', 'The'))
    [0]

    >>> list(find_all(
    ... 'Carved symbols in a mountain hollow on the bank of an inlet irritated an eccentric person',
    ... 'an'))
    [44, 51, 70]

    >>> list(find_all(
    ... 'Carved symbols in a mountain hollow on the bank of an inlet irritated an eccentric person',
    ... 'an',
    ... 50,
    ... 60))
    [51]

    :param string: the input string
    :type string: str
    :param sub: the substring
    :type sub: str
    :return: all indices in the input string
    :rtype: __generator[str]
    """
    #pylint: disable=unused-argument
    if ignore_case:
        sub = sub.lower()
        string = string.lower()
    while True:
        start = string.find(sub, start, end)
        if start == -1:
            return
        yield start
        start += len(sub)


def get_first_defined(data, keys, default_value=None):
    """
    Get the first defined key in data.
    :param data:
    :type data:
    :param keys:
    :type keys:
    :param default_value:
    :type default_value:
    :return:
    :rtype:
    """
    for key in keys:
        if key in data:
            return data[key]
    return default_value


def is_iterable(obj):
    """
    Are we being asked to look up a list of things, instead of a single thing?
    We check for the `__iter__` attribute so that this can cover types that
    don't have to be known by this module, such as NumPy arrays.

    Strings, however, should be considered as atomic values to look up, not
    iterables.

    We don't need to check for the Python 2 `unicode` type, because it doesn't
    have an `__iter__` attribute anyway.
    """
    # pylint: disable=consider-using-ternary
    return hasattr(obj, '__iter__') and not isinstance(obj, str) or isinstance(obj, GeneratorType)


def extend_safe(target, source):
    """
    Extends source list to target list only if elements doesn't exists in target list.
    :param target:
    :type target: list
    :param source:
    :type source: list
    """
    for elt in source:
        if elt not in target:
            target.append(elt)


class _Ref:
    """
    Reference for IdentitySet
    """
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value is other.value

    def __hash__(self):
        return id(self.value)


class IdentitySet(MutableSet):  # pragma: no cover
    """
    Set based on identity
    """
    def __init__(self, items=None):  # pylint: disable=super-init-not-called
        if items is None:
            items = []
        self.refs = set(map(_Ref, items))

    def __contains__(self, elem):
        return _Ref(elem) in self.refs

    def __iter__(self):
        return (ref.value for ref in self.refs)

    def __len__(self):
        return len(self.refs)

    def add(self, value):
        self.refs.add(_Ref(value))

    def discard(self, value):
        self.refs.discard(_Ref(value))

    def update(self, iterable):
        """
        Update set with iterable
        :param iterable:
        :type iterable:
        :return:
        :rtype:
        """
        for elem in iterable:
            self.add(elem)

    def __repr__(self):  # pragma: no cover
        return f"{type(self).__name__}({list(self)})"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, invalid-name, len-as-condition

from ..loose import call


def test_loose_function():

    def func(v1, v2, v3=3, v4=4):
        return v1 + v2 + v3 + v4

    assert call(func, 1, 2) == func(1, 2)
    assert call(func, 1, 2, 3, 5) == func(1, 2, 3, 5)
    assert call(func, 1, 2, v3=4, v4=5) == func(1, 2, v3=4, v4=5)
    assert call(func, 1, 2, 3, 4, 5) == func(1, 2, 3, 4)
    assert call(func, 1, 2, 3, 4, more=5) == func(1, 2, 3, 4)


def test_loose_varargs_function():
    def func(v1, v2, *args):
        return v1 + v2 + args[0] if len(args) > 0 else 3 + args[1] if len(args) > 1 else 4

    assert call(func, 1, 2) == func(1, 2)
    assert call(func, 1, 2, 3, 5) == func(1, 2, 3, 5)
    assert call(func, 1, 2, 3, 4, 5) == func(1, 2, 3, 4)


def test_loose_kwargs_function():
    def func(v1, v2, **kwargs):
        return v1 + v2 + kwargs.get('v3', 3) + kwargs.get('v4', 4)

    assert call(func, v1=1, v2=2) == func(v1=1, v2=2)
    assert call(func, v1=1, v2=2, v3=3, v4=5) == func(v1=1, v2=2, v3=3, v4=5)


def test_loose_class():
    class Dummy:
        def __init__(self, v1, v2, v3=3, v4=4):
            self.v1 = v1
            self.v2 = v2
            self.v3 = v3
            self.v4 = v4

        def call(self):
            return self.v1 + self.v2 + self.v3 + self.v4

    assert call(Dummy, 1, 2).call() == Dummy(1, 2).call()
    assert call(Dummy, 1, 2, 3, 5).call() == Dummy(1, 2, 3, 5).call()
    assert call(Dummy, 1, 2, v3=4, v4=5).call() == Dummy(1, 2, v3=4, v4=5).call()
    assert call(Dummy, 1, 2, 3, 4, 5).call() == Dummy(1, 2, 3, 4).call()
    assert call(Dummy, 1, 2, 3, 4, more=5).call() == Dummy(1, 2, 3, 4).call()


def test_loose_varargs_class():
    class Dummy:
        def __init__(self, v1, v2, *args):
            self.v1 = v1
            self.v2 = v2
            self.v3 = args[0] if len(args) > 0 else 3
            self.v4 = args[1] if len(args) > 1 else 4

        def call(self):
            return self.v1 + self.v2 + self.v3 + self.v4

    assert call(Dummy, 1, 2).call() == Dummy(1, 2).call()
    assert call(Dummy, 1, 2, 3, 5).call() == Dummy(1, 2, 3, 5).call()
    assert call(Dummy, 1, 2, 3, 4, 5).call() == Dummy(1, 2, 3, 4).call()


def test_loose_kwargs_class():
    class Dummy:
        def __init__(self, v1, v2, **kwargs):
            self.v1 = v1
            self.v2 = v2
            self.v3 = kwargs.get('v3', 3)
            self.v4 = kwargs.get('v4', 4)

        def call(self):
            return self.v1 + self.v2 + self.v3 + self.v4

    assert call(Dummy, v1=1, v2=2).call() == Dummy(v1=1, v2=2).call()
    assert call(Dummy, v1=1, v2=2, v3=3, v4=5).call() == Dummy(v1=1, v2=2, v3=3, v4=5).call()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=pointless-statement, missing-docstring, no-member, len-as-condition

from ..pattern import StringPattern, RePattern
from ..processors import ConflictSolver
from ..rules import execute_rule
from ..match import Matches


def test_conflict_1():
    input_string = "abcdefghijklmnopqrstuvwxyz"

    pattern = StringPattern("ijklmn", "kl", "abcdef", "ab", "ef", "yz")
    matches = Matches(pattern.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)

    values = [x.value for x in matches]

    assert values == ["ijklmn", "abcdef", "yz"]


def test_conflict_2():
    input_string = "abcdefghijklmnopqrstuvwxyz"

    pattern = StringPattern("ijklmn", "jklmnopqrst")
    matches = Matches(pattern.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)

    values = [x.value for x in matches]

    assert values == ["jklmnopqrst"]


def test_conflict_3():
    input_string = "abcdefghijklmnopqrstuvwxyz"

    pattern = StringPattern("ijklmnopqrst", "jklmnopqrst")
    matches = Matches(pattern.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)

    values = [x.value for x in matches]

    assert values == ["ijklmnopqrst"]


def test_conflict_4():
    input_string = "123456789"

    pattern = StringPattern("123", "456789")
    matches = Matches(pattern.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)

    values = [x.value for x in matches]
    assert values == ["123", "456789"]


def test_conflict_5():
    input_string = "123456789"

    pattern = StringPattern("123456", "789")
    matches = Matches(pattern.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)

    values = [x.value for x in matches]
    assert values == ["123456", "789"]


def test_prefer_longer_parent():
    input_string = "xxx.1x02.xxx"

    re1 = RePattern("([0-9]+)x([0-9]+)", name='prefer', children=True, formatter=int)
    re2 = RePattern("x([0-9]+)", name='skip', children=True)

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 2
    assert matches[0].value == 1
    assert matches[1].value == 2


def test_conflict_solver_1():
    input_string = "123456789"

    re1 = StringPattern("2345678", conflict_solver=lambda match, conflicting: '__default__')
    re2 = StringPattern("34567")

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "2345678"


def test_conflict_solver_2():
    input_string = "123456789"

    re1 = StringPattern("2345678", conflict_solver=lambda match, conflicting: '__default__')
    re2 = StringPattern("34567", conflict_solver=lambda match, conflicting: conflicting)

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "34567"


def test_conflict_solver_3():
    input_string = "123456789"

    re1 = StringPattern("2345678", conflict_solver=lambda match, conflicting: match)
    re2 = StringPattern("34567")

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "34567"


def test_conflict_solver_4():
    input_string = "123456789"

    re1 = StringPattern("2345678")
    re2 = StringPattern("34567", conflict_solver=lambda match, conflicting: conflicting)

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "34567"


def test_conflict_solver_5():
    input_string = "123456789"

    re1 = StringPattern("2345678", conflict_solver=lambda match, conflicting: conflicting)
    re2 = StringPattern("34567")

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "2345678"


def test_conflict_solver_6():
    input_string = "123456789"

    re1 = StringPattern("2345678")
    re2 = StringPattern("34567", conflict_solver=lambda match, conflicting: conflicting)

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "34567"


def test_conflict_solver_7():
    input_string = "102"

    re1 = StringPattern("102")
    re2 = StringPattern("02")

    matches = Matches(re2.matches(input_string))
    matches.extend(re1.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 1
    assert matches[0].value == "102"


def test_unresolved():
    input_string = "123456789"

    re1 = StringPattern("23456")
    re2 = StringPattern("34567")

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 2

    re1 = StringPattern("34567")
    re2 = StringPattern("2345678", conflict_solver=lambda match, conflicting: None)

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 2

    re1 = StringPattern("34567", conflict_solver=lambda match, conflicting: None)
    re2 = StringPattern("2345678")

    matches = Matches(re1.matches(input_string))
    matches.extend(re2.matches(input_string))

    execute_rule(ConflictSolver(), matches, None)
    assert len(matches) == 2
